/*
    Psiphon Circumvention Platform
    Copyright (C) 2009 Psiphon Inc.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "httpd.h"
#include "http_core.h"
#include "http_protocol.h"
#include "http_config.h"
#include "apr_pools.h"
#include "apr_strings.h"
#include "apr_dbd.h"
#include "mod_dbd.h"
#include "syslog.h"

#define DEBUG
#undef DEBUG

#ifdef DEBUG
#define pa_log(args...) syslog(LOG_DEBUG, args)
#else
#define pa_log(args...)  ""
#endif

#define pa_err_log(args...) syslog(LOG_ERR, args)

module AP_MODULE_DECLARE_DATA psiphon_auth_module;

typedef struct struct_pa_dir_config
{
    int need_pa;
    int check_remote_hostname;
    char *login_path;
    char *auth_cookie;
    apr_array_header_t* noauths;
} struct_pa_dir_config;

typedef struct
{
    const char* val;
} tattr;


static void *pa_create_dir_config(apr_pool_t *p, char *dummy)
{
    struct_pa_dir_config *c = (struct_pa_dir_config *)
        apr_pcalloc(p, sizeof(struct_pa_dir_config));
    c->need_pa = 0;
    c->auth_cookie = apr_pstrdup(p, "p=");
    pa_log("pa_create_dir_config(%s)", dummy);
    return (void *) c;
}

static const char *set_pa_enable(cmd_parms *parms, void *in_dconf, int arg)
{
    struct_pa_dir_config *dconf = in_dconf;
    dconf->need_pa = arg;
    pa_log("set_pa_enable(%d)", arg);
    if(dconf->need_pa)
    {
        // also checks that user has agreed to terms of usage, so if
        // not they are redirected to auth.php which drops through
        // to term agreement after user authentication
        ap_dbd_prepare (parms->server,
            "SELECT user.id, user.grp, user.xml_html, proxy.login_url "
            "FROM user JOIN proxy WHERE user.session IS NOT NULL "
            "AND user.session = %s "
            "AND !user.locked "
            "AND user.agreed_to_usage_terms <> 0 "
            "AND proxy.hostname = %s",
            "select_session");
        ap_dbd_prepare (parms->server,
            "UPDATE user SET last_activity = UNIX_TIMESTAMP() WHERE session = %s",
            "update_session");
    }
    return NULL;
}

static const char *set_pa_remote_hostname_check(cmd_parms *parms, void *in_dconf, int arg)
{
    struct_pa_dir_config *dconf = in_dconf;
    dconf->check_remote_hostname = arg;
    pa_log("set_pa_remote_hostname_check(%d)", arg);
    if(dconf->check_remote_hostname)
    {
        ap_dbd_prepare (parms->server,
            "select login_url from proxy where hostname = %s",
            "select_loginurl");
    }
    return NULL;
}

static const char *set_pa_login_path(cmd_parms *parms, void *in_dconf, const char *arg)
{
    struct_pa_dir_config *dconf = in_dconf;
    dconf->login_path = apr_pstrcat(parms->pool, arg, NULL);
    pa_log("set_pa_login_path(%s)", arg);
    return NULL;
}

static const char* set_pa_noauth(cmd_parms* parms, void* in_dconf, const char* arg)
{
    tattr* attr;
    struct_pa_dir_config* dconf = in_dconf;
    if (dconf->noauths == NULL)
        dconf->noauths = apr_array_make(parms->pool, 4, sizeof(tattr));
    attr = apr_array_push(dconf->noauths) ;
    attr->val = arg;
    return NULL ;
}


static char *del_chars(char* str, char* chars)
{
    register char *r;
    register char *p;
    register char *e = strchr(str, '\0');

    for (r = chars; *r; r++)
    {
        for (p = str; p = strchr(p, *r);)
        {
            memmove(p, p+1, e-p);
        }
    }

    return str;
}


static char *pa_bin2hex(apr_pool_t *p, char *s)
{
    char *h;
    size_t c;

    h = (char *) apr_palloc(p, strlen(s) * 2 + 1);

    for (c = 0; *(s + c); c++)
    {
        sprintf((h+(c*2)), "%02x", (unsigned char)s[c]);
    }

    return h;
}


static int pa_set_psiphon_session(request_rec *r)
{
    const char *cookies = NULL;  // cookies string from header
    const char *cookie;          // cookie pointer
    const char *cookie_e;        // end of cookie pointer
    struct_pa_dir_config *dconf;

    pa_log("pa_set_psiphon_session()");

    dconf = (struct_pa_dir_config *) ap_get_module_config(r->per_dir_config,
        &psiphon_auth_module);

    // cookies
    cookies = apr_table_get(r->headers_in, "Cookie");

    while (cookies)
    {

        pa_log("pa_set_psiphon_session(): cookies found: %s", cookies);

        cookies = del_chars(apr_pstrcat(r->pool, cookies, NULL), " \t");

        cookie = NULL;
        cookie_e = dconf->auth_cookie;

        if (!strncmp(cookies, cookie_e, strlen(cookie_e)))
        {
            cookie = cookies;
        }
        else
        {
            cookie_e = apr_pstrcat(r->pool, ";", cookie_e, NULL);
            cookie = strstr(cookies, cookie_e);
        }

        if (!cookie)
        {
            pa_log("pa_set_psiphon_session(): No session cookie found");
            break;
        }

        pa_log("pa_set_psiphon_session(): found session cookie: %s", cookie);

        cookie += strlen(cookie_e);

        for (cookie_e = cookie; (*cookie_e >= '0' && *cookie_e <='9') || (*cookie_e >= 'a' && *cookie_e <='f'); cookie_e++);

        cookie = apr_pstrndup(r->pool, cookie, cookie_e - cookie);

        pa_log("pa_set_psiphon_session(): session: %s", cookie);
        apr_table_setn(r->subprocess_env, "psiphon_session", cookie);

        break;
    }

    pa_log("end pa_set_psiphon_session()");
    return DECLINED;
}


static int pa_check_user(request_rec *r)
{

    struct_pa_dir_config *dconf;
    const char *url_prefix;
    const char *proxy_login_url = NULL;
    const char *check_hostname;
    const char *psi_session;

    int uid = 0;
    int proxy = 0;
    int gid = 0;
    int xml_html = 0;
    int i;
    
    const char *query;
    int error;
    int nrows = 0;

    apr_status_t rv;
    apr_dbd_results_t *res = NULL;
    apr_dbd_row_t *row = NULL;
    ap_dbd_t *dbd;
    apr_dbd_prepared_t *prepared_stmt;
    request_rec* rr;

    /* Say no to subrequests and internal redirects */
    if(!ap_is_initial_req(r))
    {
        return DECLINED;
    }

    pa_log("pa_check_user()");

    dconf = (struct_pa_dir_config *) ap_get_module_config(r->per_dir_config,
        &psiphon_auth_module);

    // check if PsiphonAuthEnable is On
    if (!dconf->need_pa)
    {
        pa_log("pa_check_user(): no need pa");
        return DECLINED;
    };

    // check if login page is specified
    if(!dconf->login_path || !*dconf->login_path)
    {
        pa_log("pa_check_user(): no login path specified");
        return HTTP_INTERNAL_SERVER_ERROR;
    }
    
    // check if current uri is of "noauth"
    if(dconf->noauths!= NULL)
    {
        for (i=0; i < dconf->noauths->nelts; ++i)
        {
            tattr* attrs = (tattr*)dconf->noauths->elts;
            if ( !strcmp(r->uri, attrs[i].val))
            {
                return DECLINED;
            }
        }
    }
    
    check_hostname = (const char*) ap_get_server_name(r);
    
    //check current user session

    psi_session = apr_table_get(r->subprocess_env, "psiphon_session");
    
    if (psi_session && *psi_session)
    {

        if (!(dbd = ap_dbd_acquire(r)))
        {
            pa_err_log("Can't get DB connect");
            return HTTP_INTERNAL_SERVER_ERROR;
        }

        prepared_stmt = apr_hash_get(dbd->prepared, "select_session",APR_HASH_KEY_STRING);

        if (!prepared_stmt)
        {
            pa_err_log("pa_check_user(): failed to get prepared statement 'select_session'");
            return HTTP_INTERNAL_SERVER_ERROR;
        }

        if (error = apr_dbd_pvselect(dbd->driver, r->pool, dbd->handle, &res, prepared_stmt, 0, psi_session, 
                    apr_psprintf(r->pool, "%s", check_hostname), NULL))
        {
            pa_err_log("pa_check_user(): %s", apr_dbd_error(dbd->driver, dbd->handle, error));
            return HTTP_INTERNAL_SERVER_ERROR;
        }

        for (rv = apr_dbd_get_row(dbd->driver, r->pool, res, &row, -1);
            rv != -1;
            rv = apr_dbd_get_row(dbd->driver, r->pool, res, &row, -1))
        {

            if (rv != 0)
            {
                pa_err_log("pa_check_user(): apr_dbd_get_row(): %d - %s", rv, apr_dbd_error(dbd->driver, dbd->handle, rv));
                return HTTP_INTERNAL_SERVER_ERROR;
            }

            uid = atoi(apr_dbd_get_entry(dbd->driver, row, 0));
            gid = atoi(apr_dbd_get_entry(dbd->driver, row, 1));
            xml_html = atoi(apr_dbd_get_entry(dbd->driver, row, 2));
            proxy_login_url = apr_dbd_get_entry(dbd->driver, row, 3);

            /* we can't break out here or row won't get cleaned up */
        }
        if (uid)
        {
            pa_log("pa_check_user(): user found, id %d:%d", uid, gid);

            apr_table_setn(r->subprocess_env, "psiphon_uid", apr_psprintf(r->pool, "%d", uid));
            apr_table_setn(r->subprocess_env, "psiphon_gid", apr_psprintf(r->pool, "%d", gid));
            apr_table_setn(r->subprocess_env, "psiphon_login_url", proxy_login_url);
            if (!xml_html)
            {
                apr_table_set(r->subprocess_env, "psiphon_remove_mod_transform", "1");
            }
            /*update user session time stamp and return OK*/
            prepared_stmt = apr_hash_get(dbd->prepared, "update_session",APR_HASH_KEY_STRING);

            if (!prepared_stmt)
            {
                pa_err_log("pa_check_user():failed to get prepared statement: update_session");
                return HTTP_INTERNAL_SERVER_ERROR;
            }

            if (error = apr_dbd_pvquery(dbd->driver, r->pool, dbd->handle, &nrows, prepared_stmt, psi_session, NULL))
            {
                pa_err_log("pa_check_user(): %s", apr_dbd_error(dbd->driver, dbd->handle, error));
                return HTTP_INTERNAL_SERVER_ERROR;
            }
            
            pa_log("pa_check_user(): user session has been updated");

            return DECLINED;
        }
        
        pa_log("pa_check_user(): no user found");
    }


/***********************************************************/    
    // there's no valid user session at this point,
    // get login URL for current proxy IP from DB if PsiphonLoginURLCheck is On
    if(dconf->check_remote_hostname)
    {
        if(!check_hostname)
        {
            pa_log("pa_check_user(): check_hostname == NULL");
            return HTTP_NOT_FOUND;
        } 
        
        pa_log("pa_check_user(): check_hostname == %s", check_hostname);
        
        if (!(dbd = ap_dbd_acquire(r)))
        {
            pa_err_log("Can't get DB connect");
            return HTTP_INTERNAL_SERVER_ERROR;
        }

        prepared_stmt = apr_hash_get(dbd->prepared, "select_loginurl", APR_HASH_KEY_STRING);

        if (!prepared_stmt)
        {
            pa_err_log("pa_check_user(): failed to get prepared statement: select_loginurl");
            return HTTP_INTERNAL_SERVER_ERROR;
        }

        proxy_login_url = NULL;
        row = NULL;
        res = NULL;
        
        if (error = apr_dbd_pvselect(dbd->driver, r->pool, dbd->handle, &res, prepared_stmt, 0, 
                    apr_psprintf(r->pool, "%s",  check_hostname), NULL))

        {
            pa_err_log("pa_check_user(): %s", apr_dbd_error(dbd->driver, dbd->handle, error));
            return HTTP_INTERNAL_SERVER_ERROR;
        }

        for (rv = apr_dbd_get_row(dbd->driver, r->pool, res, &row, -1);
            rv != -1;
            rv = apr_dbd_get_row(dbd->driver, r->pool, res, &row, -1))
        {

            if (rv != 0)
            {
                pa_err_log("pa_check_user(): apr_dbd_get_row(): %d - %s", rv, 
                        apr_dbd_error(dbd->driver, dbd->handle, rv));
                return HTTP_INTERNAL_SERVER_ERROR;
            }
            
            proxy_login_url = apr_dbd_get_entry(dbd->driver, row, 0);
            pa_log("pa_check_user(): proxy_login_url == %s", proxy_login_url);

        }
        
        if(!r->unparsed_uri || !*r->unparsed_uri)
        {
            pa_err_log("pa_check_user(): r->unparsed_uri is undefined");
            return HTTP_INTERNAL_SERVER_ERROR;
        }

        if(!proxy_login_url || !*proxy_login_url)
        {
            pa_log("pa_check_user(): proxy_login_url is undefined");
            return HTTP_NOT_FOUND;
        }
        
        //let's check if we've hit proxy login URL, return 404 if not
        if(strcmp(proxy_login_url, r->unparsed_uri))
        {
            pa_log("pa_check_user(): proxy_login_url !== r->unparsed_uri");
            /*
                Emulate redirect to a directory: 
                if proxy_login_url ends with '/' and r->unparsed_uri is not 
                then compare pre forward slash portion of proxy_login_url to
                r->unparsed_uri and issue a redirect to correct 
                proxy_login_url if there's a match
            */
            pa_log("pa_check_user(): proxy_login_url [strlen(proxy_login_url) - 1] == %c", 
                proxy_login_url[strlen(proxy_login_url) - 1]);

            pa_log("pa_check_user(): r->unparsed_uri == %s", r->unparsed_uri);

            if(proxy_login_url[strlen(proxy_login_url) - 1] == '/' && 
                    r->unparsed_uri[strlen(r->unparsed_uri) - 1] != '/' && 
                    strlen(r->unparsed_uri) == strlen(proxy_login_url) - 1 &&
                    0 == strncmp(proxy_login_url, r->unparsed_uri, strlen(r->unparsed_uri)))
            {
                apr_table_setn(r->headers_out, "Location", proxy_login_url);
                return HTTP_MOVED_PERMANENTLY;
            }
            /* otherwise */
            return HTTP_NOT_FOUND;
        }
    }

    //return login page
    rr = ( request_rec*)ap_sub_req_lookup_uri(dconf->login_path, r, NULL);
    if (rr->status == HTTP_OK) {
        ap_internal_fast_redirect(rr, r);
        return OK;
    }
    //dconf->login_path is invalid URL
    return HTTP_NOT_FOUND; 
}


static const command_rec dir_cmds[] =
{
    AP_INIT_FLAG("PsiphonAuthEnable", set_pa_enable, NULL, ACCESS_CONF,
        "Enable psiphon authorization (on/off)"),
    AP_INIT_FLAG("PsiphonLoginURLCheck", set_pa_remote_hostname_check, NULL, ACCESS_CONF,
        "Enable psiphon authorization login URL per remote hostname check (on/off)"),
    AP_INIT_TAKE1("PsiphonAuthLoginPath", set_pa_login_path, NULL, ACCESS_CONF,
        "URL of the login script"),
    AP_INIT_ITERATE("PsiphonNoAuth", set_pa_noauth, NULL, ACCESS_CONF,
        "URLs that don't require authorization"),
    {NULL}
};

static void pa_hooks(apr_pool_t *pool)
{
    static const char *aSucc[] = { "mod_map_to_proxy.c", NULL };
    ap_hook_post_read_request(pa_set_psiphon_session, NULL, aSucc, APR_HOOK_FIRST);
    ap_hook_access_checker(pa_check_user, NULL, NULL, APR_HOOK_LAST);
}


module AP_MODULE_DECLARE_DATA psiphon_auth_module =
{
    STANDARD20_MODULE_STUFF,
    pa_create_dir_config,
    NULL,
    NULL,
    NULL,
    dir_cmds,
    pa_hooks
};
