/*
    Psiphon Circumvention Platform
    Copyright (C) 2009 Psiphon Inc.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "apr.h"
#include "apr_strings.h"
#include "apr_thread_proc.h"
#include "apr_hash.h"
#include "apr_user.h"
#include "apr_lib.h"
#include "apr_optional.h"

#define APR_WANT_STRFUNC
#define APR_WANT_MEMFUNC
#include "apr_want.h"

#include "ap_config.h"
#include "util_filter.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_request.h"
#include "http_core.h"
#include "http_protocol.h"
#include "http_log.h"
#include "http_main.h"
#include "util_script.h"
#include "http_core.h"

#define DEBUG
#undef DEBUG

#ifdef DEBUG
#define bluebar_log(args...) syslog(LOG_DEBUG, args)
#else
#define bluebar_log(args...)  ""
#endif

module AP_MODULE_DECLARE_DATA bluebar_module;

typedef struct bluebar_conf
{
    const char *bluebar_uri;
} bluebar_conf;

typedef struct bluebar_input_ctx
{
    ap_filter_t *ff;
}


bluebar_input_ctx;

static const char *add_bluebar_uri(cmd_parms *cmd, void *CFG, const char *arg)
{
    bluebar_conf *cfg = CFG;
    cfg->bluebar_uri = apr_pstrdup(cmd->pool, arg);
    return NULL;
}


static void *create_bluebar_config(apr_pool_t *p, char *dummy)
{
    bluebar_conf *cfg =
        (bluebar_conf *) apr_pcalloc(p, sizeof(bluebar_conf));

    cfg->bluebar_uri = NULL;
    return (void *) cfg;
}


static apr_status_t bluebar_filter(ap_filter_t *f, apr_bucket_brigade *bb)
{
    bluebar_input_ctx *input_ctx;
    apr_bucket *b;
    apr_bucket_brigade *end;
    request_rec *rr;
    apr_pool_t* subpool;
    apr_status_t rv;
    
    int seen_eos = 0;

    request_rec * r = f->r;
    bluebar_conf *conf = ap_get_module_config(r->per_dir_config,
        &bluebar_module);

    if (f->r->main != NULL)  /*no subrequests*/
    {
        ap_remove_output_filter(f);
        return ap_pass_brigade(f->next, bb);
    }

    if ( strncasecmp(r->content_type, "text/html", 9) &&
            strncasecmp(r->content_type, "application/xhtml+xml", 21))
    {
        ap_remove_output_filter(f) ;
        return ap_pass_brigade(f->next, bb);
    }


    apr_table_unset(f->r->headers_out, "Content-Length");
    apr_table_unset(f->r->headers_out, "ETag");

    if(!conf->bluebar_uri)
    {
        ap_remove_output_filter(f) ;
        return ap_pass_brigade(f->next, bb);
    }

    rr = ap_sub_req_lookup_uri(conf->bluebar_uri , r, NULL);

    if (rr->status != HTTP_OK)
    {
        ap_destroy_sub_req(rr);
        ap_remove_output_filter(f) ;
        return ap_pass_brigade(f->next, bb);
    }

    input_ctx = apr_palloc(r->pool, sizeof(input_ctx));
    input_ctx->ff = f->next;

    ap_add_output_filter("bluebar-store",  input_ctx, rr, r->connection);

    for (b = APR_BRIGADE_FIRST(bb);
         b != APR_BRIGADE_SENTINEL(bb);
         b = APR_BUCKET_NEXT(b))
    {
        if (APR_BUCKET_IS_EOS(b) && !seen_eos)
        {
            seen_eos = 1;
            apr_bucket_brigade *end;

            end = apr_brigade_split(bb, b);
            if (!APR_BRIGADE_EMPTY(bb))
            {
                ap_pass_brigade(f->next, bb);
            }
            apr_brigade_cleanup(bb);
            bb = end;
            ap_run_sub_req(rr);
            ap_destroy_sub_req(rr);
            continue;
        }
        if(seen_eos)
        {
            APR_BUCKET_REMOVE(b);
            apr_bucket_destroy(b);
        }
    }
    rv = ap_pass_brigade(f->next, bb);
    apr_brigade_cleanup(bb);
    return rv;
}


apr_status_t bluebar_store_filter(ap_filter_t * f, apr_bucket_brigade * bb)
{
    apr_status_t rv;
    apr_bucket *b;
    bluebar_input_ctx *ctxt = f->ctx;
    int seen_eos = 0;
    if (APR_BRIGADE_EMPTY(bb))
    {
        apr_brigade_cleanup(bb);
        return APR_SUCCESS;
    }
    for (b = APR_BRIGADE_FIRST(bb);
         b != APR_BRIGADE_SENTINEL(bb);
         b = APR_BUCKET_NEXT(b))
    {
        if (APR_BUCKET_IS_EOS(b)) /*shouldn't pass subrequest's eos?*/
        {
            APR_BUCKET_REMOVE(b);
            apr_bucket_destroy(b);
        }
    }
    rv = ap_pass_brigade(ctxt->ff, bb);
    apr_brigade_cleanup(bb);
    return rv;
}


static const command_rec bluebar_cmds[] =
{
    AP_INIT_TAKE1("BluebarURI", add_bluebar_uri, NULL, ACCESS_CONF || OR_FILEINFO,
    "Bluebar URI"),
    {NULL}
};

static void bluebar_register_hook(apr_pool_t *p)
{
    ap_register_output_filter_protocol("psiphon-bluebar", bluebar_filter,
        NULL, AP_FTYPE_RESOURCE,
        AP_FILTER_PROTO_CHANGE|AP_FILTER_PROTO_CHANGE_LENGTH) ;
    ap_register_output_filter_protocol("bluebar-store", bluebar_store_filter,
        NULL, AP_FTYPE_RESOURCE,
        AP_FILTER_PROTO_CHANGE|AP_FILTER_PROTO_CHANGE_LENGTH) ;

}


module AP_MODULE_DECLARE_DATA bluebar_module =
{
    STANDARD20_MODULE_STUFF,
    create_bluebar_config,       /* create per-directory config structure */
    NULL,                        /* merge per-directory config structures */
    NULL,                        /* create per-server config structure */
    NULL,                        /* merge per-server config structures */
    bluebar_cmds,                /* command apr_table_t */
    bluebar_register_hook        /* register hooks */
};
