/* Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "apr.h"
#include "apr_strings.h"

#if APR_HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_log.h"
#include "apr_optional.h"
#include "mod_post2get.h"
#include "mod_proxy_html.h"
#include "mod_proxy_js.h"
#include "mod_proxy_css.h"
#include "mod_proxy_flash.h"

module AP_MODULE_DECLARE_DATA psiphon_env_module;

static APR_OPTIONAL_FN_TYPE(psiphon_get_js_prefix) *get_js_prefix_fn = NULL;
static APR_OPTIONAL_FN_TYPE(psiphon_get_html_prefix) *get_html_prefix_fn = NULL;
static APR_OPTIONAL_FN_TYPE(psiphon_get_post2get_prefix) *get_frm_prefix_fn = NULL;
static APR_OPTIONAL_FN_TYPE(psiphon_get_css_prefix) *get_css_prefix_fn = NULL;
static APR_OPTIONAL_FN_TYPE(psiphon_get_flash_prefix) *get_flash_prefix_fn = NULL;



static int psiphon_fixup_env(request_rec *r)
{
    
    const char* html_prefix = NULL;
    const char* frm_prefix = NULL;
    const char* js_prefix = NULL;
    const char* css_prefix = NULL;
    const char* flash_prefix = NULL;
    if(get_html_prefix_fn == NULL)
    {
        get_html_prefix_fn = APR_RETRIEVE_OPTIONAL_FN(psiphon_get_html_prefix);
        if(get_html_prefix_fn == NULL)
        {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "mod_proxy_html is not loaded.") ;
            return HTTP_INTERNAL_SERVER_ERROR;
        }
    }
    if(get_frm_prefix_fn == NULL)
    {
        get_frm_prefix_fn = APR_RETRIEVE_OPTIONAL_FN(psiphon_get_post2get_prefix);
        if(get_frm_prefix_fn == NULL)
        {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "mod_post2get is not loaded.") ;
            return HTTP_INTERNAL_SERVER_ERROR;
        }
    }
    if(get_js_prefix_fn == NULL)
    {
        get_js_prefix_fn = APR_RETRIEVE_OPTIONAL_FN(psiphon_get_js_prefix);
        if(get_js_prefix_fn == NULL)
        {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "mod_proxy_js is not loaded.") ;
            return HTTP_INTERNAL_SERVER_ERROR;
        }
    }
    if(get_css_prefix_fn == NULL)
    {
        get_css_prefix_fn = APR_RETRIEVE_OPTIONAL_FN(psiphon_get_css_prefix);
        if(get_css_prefix_fn == NULL)
        {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "mod_proxy_css is not loaded.") ;
            return HTTP_INTERNAL_SERVER_ERROR;
        }
    }
    if(get_flash_prefix_fn == NULL)
    {
        get_flash_prefix_fn = APR_RETRIEVE_OPTIONAL_FN(psiphon_get_flash_prefix);
        if(get_flash_prefix_fn == NULL)
        {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "mod_proxy_flash is not loaded.") ;
            return HTTP_INTERNAL_SERVER_ERROR;
        }
    }
    
    html_prefix = get_html_prefix_fn(r);
    
    if(!html_prefix || !*html_prefix)
    {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "html prefix is empty.") ;
        return HTTP_INTERNAL_SERVER_ERROR;
    
    }
    
    frm_prefix = get_frm_prefix_fn(r);
    
    if(!frm_prefix || !*frm_prefix)
    {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "form prefix is empty.") ;
        return HTTP_INTERNAL_SERVER_ERROR;
    
    }

    js_prefix = get_js_prefix_fn(r);
    
    if(!js_prefix || !*js_prefix)
    {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "javascript prefix is empty.") ;
        return HTTP_INTERNAL_SERVER_ERROR;
    
    }

    css_prefix = get_css_prefix_fn(r);
    
    if(!css_prefix || !*css_prefix)
    {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "css prefix is empty.") ;
        return HTTP_INTERNAL_SERVER_ERROR;
    
    }

    flash_prefix = get_flash_prefix_fn(r);
    
    if(!flash_prefix || !flash_prefix)
    {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "flash prefix is empty.") ;
        return HTTP_INTERNAL_SERVER_ERROR;
    
    }
    apr_table_setn(r->subprocess_env, "psiphon_html_prefix", html_prefix);
    apr_table_setn(r->subprocess_env, "psiphon_frm_prefix", frm_prefix);
    apr_table_setn(r->subprocess_env, "psiphon_js_prefix", js_prefix);
    apr_table_setn(r->subprocess_env, "psiphon_css_prefix", css_prefix);
    apr_table_setn(r->subprocess_env, "psiphon_flash_prefix", flash_prefix);
    return OK;
}

static void register_hooks(apr_pool_t *p)
{
    ap_hook_post_read_request(psiphon_fixup_env, NULL, NULL, APR_HOOK_FIRST);
//    ap_hook_fixups(psiphon_fixup_env, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA psiphon_env_module =
{
    STANDARD20_MODULE_STUFF,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    register_hooks
};
