/*
    Psiphon Circumvention Platform
    Copyright (C) 2009 Psiphon Inc.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "mod_psiphon_headers.h"

#define PSIPHON_HEADERS_FNAME "psiphon-headers-out"

// maximum allowed number of dots in domain name we want to handle
#define MAX_DOTS_PER_DOMAIN 10

// for cookies specs
// SEE http://web.archive.org/web/20070805052634/http://wp.netscape.com/newsref/std/cookie_spec.html

// maximum allowed cookies per domain
#define MAX_COOKIES_PER_DOMAIN 20
// maximum cookie size
#define MAX_COOKIE_LINE 4096

module AP_MODULE_DECLARE_DATA psiphon_headers_module ;

typedef struct psiphon_headers_conf
{
    int psiphon_headers_enable ;
    const char* loc ;
} psiphon_headers_conf ;

typedef struct psiphon_headers_ctxt
{
    int uid ;
    const char* uri ;
} psiphon_headers_ctxt ;

static psiphon_headers_ctxt* create_context(request_rec* r)
{
    int uid;
    const char* str_uid;
    char* proxy_uri;
    psiphon_headers_ctxt* ctxt;

    /*Get authenticated user id and proxy uri*/
    str_uid =(const char*) apr_table_get(r->subprocess_env, "psiphon_uid");

    if(!str_uid || !*str_uid)
    {
        return NULL;
    }

    uid = atoi(str_uid);

    if(!uid)
    {
        return NULL;
    }
    proxy_uri = &r->filename[6];

    if(!proxy_uri || !*proxy_uri)
    {
        return NULL;
    }

    ctxt =(psiphon_headers_ctxt *) apr_pcalloc(r->pool, sizeof(psiphon_headers_ctxt));

    ctxt->uid = uid;
    ctxt->uri =(const char*) apr_pstrdup(r->pool, proxy_uri);

    return ctxt;
}


static void fixup_referer(request_rec* r, psiphon_headers_ctxt* ctx)
{

    const char* ref;
    int rv;
    apr_uri_t uri;
    const char* uri_path;
    const char* server_scheme;
    const char* server_hostname;
    psiphon_headers_conf * conf;

    conf =(psiphon_headers_conf *) ap_get_module_config(r->per_dir_config,
        &psiphon_headers_module);

    server_scheme =(const char*) ap_http_scheme(r);
    server_hostname =(const char*) ap_get_server_name(r);

    if(!r->headers_in)
    {
        /*no input headers*/
        return;
    }

    ref = apr_table_get(r->headers_in, "Referer");

    if(!ref || !*ref)
    {
        /*no referer value, unset*/
        apr_table_unset(r->headers_in, "Referer");
        return;
    }
    rv = apr_uri_parse(r->pool, ref, &uri);

    if(rv != APR_SUCCESS)
    {
        /*unparsable referer*/
        return;
    }
    if(!uri.hostname || !uri.scheme || !uri.path)
    {
        /*Missing hostname or scheme or path*/
        return;
    }
    if(strcasecmp(uri.scheme, server_scheme)
        || strcasecmp(uri.hostname, server_hostname))
    {
        /*Refered not by us*/
        return;
    }
    if(uri.query && *uri.query)
    {
        uri_path = apr_psprintf(r->pool, "%s?%s", uri.path, uri.query);
    }
    else
    {
        uri_path = uri.path;
    }
    if(!conf->loc || !*conf->loc
        || strncmp(conf->loc, uri_path, strlen(conf->loc)))
    {
        /*Refered by us but prefix doesn't match*/
        return;
    }

    uri_path +=(strlen(conf->loc));

    if(uri_path[0] != '/')
    {
        /*no trailing '/' after prefix*/
        return;
    }

    uri_path++;  /*eat '/'*/

    if(uri_path && *uri_path)
    {
        psiphon_headers_log("new Referer = %s", uri_path);
        apr_table_set(r->headers_in, "Referer", uri_path);
        return;
    }
    /*empty URL, unset*/
    apr_table_unset(r->headers_in, "Referer");

}


static const char* psiphon_proxify_location(request_rec* r, const char* value, psiphon_headers_ctxt* ctx)
{
    const char* uri;
    const char* return_url;
    apr_pool_t *p = r->pool;
    unsigned port = ap_get_server_port(r);
    const char* hostname =(const char*) ap_get_server_name(r);

    psiphon_headers_conf * conf;
    conf =(psiphon_headers_conf *) ap_get_module_config(r->per_dir_config,
        &psiphon_headers_module);

    psiphon_headers_log("conf->loc = %s, base_url = %s", conf->loc, ctx->uri);

    uri =(const char*) psi_proxify_url(p, value, ctx->uri);

    if(ap_is_default_port(port, r))
    {
        return_url=(const char*) apr_psprintf(p, "%s://%s%s/%s", ap_http_scheme(r),
            hostname, conf->loc, uri);
    }
    else
    {
        return_url=(const char*) apr_psprintf(p, "%s://%s:%u%s/%s", ap_http_scheme(r),
            hostname, port, conf->loc, uri);
    }

    return return_url;
}


static int psiphon_db_to_cookie(request_rec *r, psiphon_headers_ctxt* ctx)
{
    const char *cookies = NULL;  // cookies string from header
    const char *cookie;          // cookie pointer
    const char *cookie_e;        // end of cookie pointer

    struct apr_uri_t *uri = apr_pcalloc(r->pool, sizeof(apr_uri_t));

    apr_status_t rv;
    apr_dbd_results_t *res = NULL;
    apr_dbd_row_t *row = NULL;
    ap_dbd_t *dbd;
    apr_dbd_prepared_t *prepared_stmt;
    char* pstmt_key;
    int error;

    char *domain;
    char *now_param; 
    char *uid_param; 

    apr_hash_t *cookies_table;
    apr_array_header_t *domains_array;
    char **domain_ptr;

    char *tmp_cl;                // temporary cookie line
    char *cl;                    // cookies line
    apr_hash_index_t *hi;        // hash table index
    void *val;
    char *value;                 // hash table value pointer
    char *key;                   // hash table key pointer

    int i;
    int dot_count = 0;


    time_t t = time(NULL);

    now_param = apr_psprintf(r->pool, "%d",(int) t);
    uid_param = apr_psprintf(r->pool, "%d", ctx->uid);

    if(apr_uri_parse(r->pool, ctx->uri, uri) != APR_SUCCESS)
    {
        psiphon_headers_log("Can't parse %s", ctx->uri);
        return FALSE;
    }

    if(strcmp(uri->scheme, "http") && strcmp(uri->scheme, "https"))
    {
        psiphon_headers_log("%s scheme skipped, continue", uri->scheme); 
        return TRUE;
    }

    if(!uri->hostname || !strlen(uri->hostname))
    {
        psiphon_headers_log("uri->hostname is empty, break");
        return FALSE;
    }

    if(!uri->path || !strlen(uri->path))
    {
        psiphon_headers_log("uri->path is empty, setting to /");
        uri->path = apr_pstrdup(r->pool, "/");
    }

    psiphon_headers_log("got %s :// %s %s", uri->scheme, uri->hostname, uri->path);

    domain = uri->hostname;

    //Count dots in domain name
    for (i = 0; i < strlen(domain); ++i)
        if (domain[i] == '.')
            ++dot_count;

    // There is more dots in the domain name than we want to handle
    if(dot_count > MAX_DOTS_PER_DOMAIN)
    {
        psiphon_headers_err_log("too many dots in domain name");
        return FALSE;
    }

    /*    
    Make a table of all possible matching cookie domains for given hostname
    i.e. if hostname is 'aa.bb.cc' then domains_array will 
    hold '.aa.bb.cc', 'aa.bb.cc' and '.bb.cc' and '.cc' will
    be omitted by cookie_valid_domain() 
    */
    domains_array = apr_array_make(r->pool, dot_count + 2, sizeof(char*)); //+2 to hold 'domain' itself and '.domain' 

    //Insert '.domain' and domain
    domain_ptr = apr_array_push(domains_array);
    *domain_ptr = apr_psprintf(r->pool, ".%s", domain);

    domain_ptr = apr_array_push(domains_array);
    *domain_ptr = apr_pstrdup(r->pool, domain);

    for(domain; domain = strchr(++domain, '.');)
    {
        if(cookie_valid_domain(domain, uri->hostname))
        {
            domain_ptr = apr_array_push(domains_array);
            *domain_ptr = apr_pstrdup(r->pool, domain);
        }
    }

    //define which prepared statement to use(https vs http)
    if(!strcmp(uri->scheme, "http"))
    {
        pstmt_key = "psiphon_select_http_cookie";
    }
    else
    {
        pstmt_key = "psiphon_select_https_cookie";
    }

    if(!(dbd = ap_dbd_acquire(r)))
    {
        psiphon_headers_err_log("Can't get DB connection");
        return FALSE;
    }

    prepared_stmt = apr_hash_get(dbd->prepared, pstmt_key, APR_HASH_KEY_STRING);
    if(!prepared_stmt)
    {
        psiphon_headers_err_log("failed to get prepared statement: %s", pstmt_key);
        return FALSE;
    }

    cookies_table = apr_hash_make(r->pool);

    domain_ptr = (char **) domains_array->elts;
    for(i = 0; i < domains_array->nelts; i++)
    {
        domain = domain_ptr[i];

        if(error = apr_dbd_pvselect(dbd->driver, r->pool, dbd->handle, &res,
            prepared_stmt,  0, uid_param, domain, now_param))
        {
            //DB error occured
            psiphon_headers_err_log("%s", apr_dbd_error(dbd->driver, dbd->handle, error));
            return FALSE;
        }
        for(rv = apr_dbd_get_row(dbd->driver, r->pool, res, &row, -1);
            rv != -1;
            rv = apr_dbd_get_row(dbd->driver, r->pool, res, &row, -1))
        {

            if(rv != 0)
            {
                //DB error occured
                psiphon_headers_err_log("apr_dbd_get_row(): %d - %s", rv, apr_dbd_error(dbd->driver, dbd->handle, rv));
                return FALSE;
            }

            //per cookies spec max # of cookies per domain shouldn't exceed 20
            if(apr_hash_count(cookies_table) <= MAX_COOKIES_PER_DOMAIN)
            {

                if(!strncmp(apr_dbd_get_entry(dbd->driver, row, 1), uri->path, strlen(apr_dbd_get_entry(dbd->driver, row, 1))))
                {

                    key = apr_pstrcat(r->pool, apr_dbd_get_entry(dbd->driver, row, 0), ";", //hash key == 'name;path'
                                      apr_dbd_get_entry(dbd->driver, row, 1), NULL);

                    if(!apr_hash_get(cookies_table, key, APR_HASH_KEY_STRING))
                    {
                    /*
                        in case we have a cookie foo=bar; path=/ for www.domain.com
                        and another one foo=foobar;path=/ for .domain.com the most
                        recent and having longest domain in the SQL table takes 
                        precedence
                    */
                        value = apr_pstrdup(r->pool, apr_dbd_get_entry(dbd->driver, row, 2));  //hash value = 'name=value'
                        psiphon_headers_log("adding \"%s\" => \"%s\"", key, value);
                        if(value)
                        {
                            apr_hash_set(cookies_table, key, APR_HASH_KEY_STRING, value);
                        }
                    }
                    else 
                    {
                        psiphon_headers_log("\"%s\" exists", key);
                    }
                }
                else
                {
                    psiphon_headers_log("\"%s\" does not include \"%s\"", uri->path, apr_dbd_get_entry(dbd->driver, row, 1));
                }
            }
            else
            {
                psiphon_headers_log("cookie table is full");
                /* we can't break out here or row won't get cleaned up */
            }
        }
    }

    cl = NULL;

    for(hi = apr_hash_first(r->pool, cookies_table); hi; hi = apr_hash_next(hi))
    {
        apr_hash_this(hi, NULL, NULL, &val);

        psiphon_headers_log("\"%s\"",(char*) val);

        if(val)
        {
            if(cl)
                tmp_cl = apr_pstrcat(r->pool, cl,(char*) val, "; ", NULL);
            else
                tmp_cl = apr_psprintf(r->pool, "%s; ",(char*) val);

            //Cookie line is  too long as per spec, break out.
            if(tmp_cl && strlen(tmp_cl) > MAX_COOKIE_LINE)
                break;

            cl = tmp_cl;
        }
    }

    if(cl)
    {
        psiphon_headers_log("Cookie: %s", cl);
        apr_table_setn(r->headers_in, "Cookie", cl);
    }

    return TRUE;
}


static int psiphon_cookie_to_db(request_rec *r, const char *val, psiphon_headers_ctxt* ctx)
{
    // database stuff
    apr_status_t rv;
    apr_dbd_results_t *res = NULL;
    apr_dbd_row_t *row = NULL;
    ap_dbd_t *dbd;
    apr_dbd_prepared_t *prepared_stmt;
    int error;
    int nrows;
    //vars to hold string representation of integers
    char* expires_param;
    char* uid_param;
    char* secure_param;

    // cookies stuff
    const char* domain;
    const char* path;
    const char* name;
    const char* value;
    const char* expires_str;
    time_t expires;
    int secure;

    char *pb;
    char *pe;
    char *str;
    size_t len;

    struct apr_uri_t *uri = apr_pcalloc(r->pool, sizeof(apr_uri_t));

    time_t t = time(NULL);
    char *tstr;

    tstr = apr_psprintf(r->pool, "%d",(int) t);

    if(!(dbd = ap_dbd_acquire(r)))
    {
        psiphon_headers_err_log("Can't get DB connect");
        return FALSE;
    }

    if(apr_uri_parse(r->pool, ctx->uri, uri) != APR_SUCCESS)
    {
        psiphon_headers_log("Can't parse %s", ctx->uri);
        return FALSE;
    }

    if(strcmp(uri->scheme, "http") && strcmp(uri->scheme, "https"))
    {
        psiphon_headers_log("%s scheme skipped", uri->scheme);
        return FALSE;
    }

    if(!uri->hostname || !strlen(uri->hostname))
    {
        psiphon_headers_log("uri->hostname is empty");
        return FALSE;
    }

    if(!uri->path || !strlen(uri->path))
    {
        psiphon_headers_log("uri->path is empty, setting to /");
        uri->path = apr_pstrcat(r->pool, "/", NULL);
    }

    psiphon_headers_log("got %s :// %s %s", uri->scheme, uri->hostname, uri->path);

    // reading cookies

    str = apr_pstrcat(r->pool, ";", val, ";", NULL);

    // cleaning stuff aroung each ';'
    for(pb = strchr(str, ';'); pb = strchr(pb, ';'); pb++)
    {

        // remove spaces before current pointer
        while(((pb-1) > str) &&(*(pb-1) == ' ' || *(pb-1) == '\t'))
        {
            memmove(pb-1, pb, strlen(pb) +1);
            pb--;
        }

        // remove spaces after current pointer
        while(*(pb+1) == ' ' || *(pb+1) == '\t')
        {
            memmove(pb+1, pb+2, strlen(pb+2) +1);
        }
    }

    // get domain ///////////
    while(1)
    {
        if(!(pb = ap_strcasestr(str, ";domain=")))
        {
            domain = apr_pstrdup(r->pool, uri->hostname);
            break;
        }
        pb += 8;
        for(pe = pb; *pe && *pe != ';'; pe++);
        domain = apr_pstrndup(r->pool, pb, pe-pb);

        memmove(pb-8, pe , strlen(pe) +1);

        if(!cookie_valid_domain(domain, uri->hostname))
        {
            //malicious cookie, invalid domain, don't store, return TRUE;
            psiphon_headers_log("%s not a valid domain for %s", domain, ctx->uri);
            return TRUE;
        }
        break;
    }

    // get path ///////////
    while(1)
    {
        if(!(pb = ap_strcasestr(str, ";path=")))
        {
            path = apr_pstrdup(r->pool, uri->path);
            break;
        }
        pb += 6;
        for(pe = pb; *pe && *pe != ';'; pe++);
        path = apr_pstrndup(r->pool, pb, pe-pb);

        memmove(pb-6, pe , strlen(pe) +1);

        break;
    }

    // get expires ///////////
    while(1)
    {
        if(!(pb = ap_strcasestr(str, ";expires=")))
        {
            expires = -1;
            expires_str = "-1";
            break;
        }
        pb += 9;
        for(pe = pb; *pe && *pe != ';'; pe++);

        expires_str = apr_pstrndup(r->pool, pb, pe-pb);

        memmove(pb-9, pe , strlen(pe) +1);

        expires = cookie_get_date(expires_str);

        break;
    }

    // get secure ///////////
    if(pb = ap_strcasestr(str, ";secure;"))
    {
        secure=1;
        memmove(pb, pb+7, strlen(pb+7) +1);
    }
    else
    {
        secure=0;
    }

    // getting cookie name

    // locating first '='
    if(!(pb = strchr(str, '=')))
    {
        psiphon_headers_log("no '=' found: %s", str);
        return TRUE;             //bad cookie, don't store, return TRUE;
    }

    // getting name
    // go back to find the begining of the name
    for(pe = pb;(pe >= str) &&(*pe != ';'); pe--);
    pe++;
    name = apr_pstrndup(r->pool, pe, pb - pe);

    // getting value
    // go forward to find the end of the name
    pb++;

    for(pe = pb;(*pe != ';') && *pe; pe++);
    value = apr_pstrndup(r->pool, pb, pe - pb);

    psiphon_headers_log("%s=%s; domain=%s; path=%s; expires=%ld; secure=%d; now_time = %ld",
        name, value, domain, path, expires, secure, t);

    nrows = 0;

    uid_param = apr_psprintf(r->pool, "%d", ctx->uid);
    expires_param = apr_psprintf(r->pool, "%d",(int) expires);
    secure_param = apr_psprintf(r->pool, "%d", secure);

    /*
        Insert/update if cookie expires in future or no expiry 
        date given(good for current session only)
    */
    if((expires > t) ||(expires == -1)) 
    {
        prepared_stmt = apr_hash_get(dbd->prepared, "psiphon_insert_cookie",APR_HASH_KEY_STRING);
        if(!prepared_stmt)
        {
            psiphon_headers_err_log("failed to get prepared statement: psiphon_insert_cookie");
            return FALSE;
        }
        if(error=apr_dbd_pvquery(dbd->driver, r->pool, dbd->handle, &nrows, prepared_stmt, 
            uid_param, domain, path, name, value, expires_param, secure_param, 
            name, value, expires_param, secure_param, uid_param, domain, path, NULL))
        {
                //DB error occurred
            psiphon_headers_err_log("DB error %s in psiphon_insert_cookie", 
                                    apr_dbd_error(dbd->driver, dbd->handle, error));
            return FALSE;
        }
        psiphon_headers_log("DB INSERT cookie %s=%s; domain=%s; path=%s; expires=%ld; secure=%d;",
                name, value, domain, path, expires, secure);
    }
    else /*Delete the cookie if any*/
    {
        prepared_stmt = apr_hash_get(dbd->prepared, "psiphon_delete_cookie",APR_HASH_KEY_STRING);
        if(!prepared_stmt)
        {
            psiphon_headers_err_log("failed to get prepared statement: psiphon_delete_cookie");
            return FALSE;
        }
        if(error=apr_dbd_pvquery(dbd->driver, r->pool, dbd->handle, &nrows, prepared_stmt, 
           uid_param, domain, path, name, NULL))
        {
            //DB error occurred
            psiphon_headers_err_log("DB error %s in psiphon_delete_cookie", 
                apr_dbd_error(dbd->driver, dbd->handle, error));
            return FALSE;
        }
        psiphon_headers_log("DB DELETE cookie %s=%s; domain=%s; path=%s; expires=%ld; secure=%d;",
                name, value, domain, path, expires, secure);
    }
    return TRUE; //All good
}


int log_header(void* p, const char* key, const char* value)
{
    char *log_str =(char*) p;
    psiphon_headers_log("headers %s: %s => %s", log_str, key, value);
    return 1;
}


int fixup_output_headers(void* f, const char* key, const char* value)
{
    ap_filter_t* ff =(ap_filter_t*) f;
    request_rec* r =(request_rec*) ff->r;
    psiphon_headers_ctxt *ctx =(psiphon_headers_ctxt*) ff->ctx;
    int i;
    static const char* uri_hdrs[] =
        { "Location", "Content-Location", "URI",  "Destination", NULL };

    if(!key || !*key || !value || !*value)
    {
        return TRUE;
    }

    /* Store cookies in DB*/
    if(!strcasecmp("Set-Cookie", key))
    {
        return psiphon_cookie_to_db(r, value, ctx);
    }
    /*Proxify URI headers*/
    for(i = 0 ; uri_hdrs[i] ; ++i)
    {
        if(!strcasecmp(uri_hdrs[i], key))
        {
            apr_table_set(r->headers_out, key,
                psiphon_proxify_location(r, value, ctx)) ;
        }
    }
    return TRUE;
}

static apr_status_t output_headers_filter(ap_filter_t* f, apr_bucket_brigade* bb)
{

    psiphon_headers_ctxt *ctx =(psiphon_headers_ctxt*) f->ctx;
    int ret = FALSE;

    ret = apr_table_do(fixup_output_headers, f, f->r->headers_out, NULL);

    if(!ret) //fixup returned FALSE, must be psiphon_cookie_to_db error
    {
        psiphon_headers_err_log("fixup_output_headers error for uri=%s", ctx->uri);
        return APR_EGENERAL;
    }

    /*  Assuming that cookies are stored in DB unset 'em*/
    apr_table_unset(f->r->headers_out, "Set-Cookie");

    /*Log headers*/
    apr_table_do(log_header, "output", f->r->headers_out, NULL);

    ap_remove_output_filter(f);
    return ap_pass_brigade(f->next,bb);
}


const char* get_user_agent_for_this_request(request_rec *r)
{
    int match_length = 0;
    int match_length_so_far = 0;
    const char* user_agent_string = 0;
    const char* domain = 0;
    const char* domain_db = 0;
    const char* path = 0;
    const char* path_db = 0;
    const char* back_link = 0;
    char* query = 0;
    apr_dbd_results_t* res = 0;
    apr_dbd_row_t* row = 0;
    ap_dbd_t* dbd = 0;
    apr_status_t error;

    apr_pool_t* p = r->pool;

    /* getting domain information */
    if (!(domain = get_request_domain(r)))
    {
        psiphon_headers_err_log("get_user_agent_for_this_request(): no r->subprocess_env->psiphon_proxy_site found");
        return 0;
    }

    /* getting the path */
    path = get_request_path(r);
    psiphon_headers_log("get_user_agent_for_this_request(): path: %s", path);

    /* getting useragent table */
    if (!(dbd = ap_dbd_acquire(r)))
    {
        psiphon_headers_err_log("get_user_agent_for_this_request(): Can't get DB connection");
        return 0;
    }

    query = apr_pstrdup(p, "SELECT domain, path, user_agent_string, back_link FROM useragent ORDER BY id DESC");

    if (error = apr_dbd_select(dbd->driver,p,dbd->handle, &res, query, 0))
    {
        psiphon_headers_err_log("get_user_agent_for_this_request(): %s: %s", apr_dbd_error(dbd->driver, dbd->handle, error), query);
        return 0;
    }

    /* iterate through the custom user agents, and find the longest one that matches domain and path */
    for (error = apr_dbd_get_row(dbd->driver, r->pool, res, &row, -1);
         error != -1;
         error = apr_dbd_get_row(dbd->driver, r->pool, res, &row, -1))     
    {
        if (error != 0)
        {
            psiphon_headers_err_log("get_user_agent_for_this_request(): apr_dbd_get_row(): %d - %s", error, apr_dbd_error(dbd->driver, dbd->handle, error));
            break;
        }

        domain_db = apr_pstrdup(p, apr_dbd_get_entry(dbd->driver, row, 0));
        path_db = apr_pstrdup(p, apr_dbd_get_entry(dbd->driver, row, 1));

        if (!request_matches_url_rule(domain, path, domain_db, path_db))
        {
            psiphon_headers_log("get_user_agent_for_this_request(): \"%s%s\" does not match \"%s%s\"", domain_db, path_db, domain, path);
            continue;
        }
        
        psiphon_headers_log("get_user_agent_for_this_request(): \"%s%s\" is in \"%s%s\"", domain_db, path_db, domain, path);

        /* assign only if longest match so far */
        match_length = strlen(path_db);
        if (match_length >= match_length_so_far)
        {
            match_length_so_far = match_length;
            user_agent_string = apr_pstrdup(p, apr_dbd_get_entry(dbd->driver, row, 2));
            back_link = apr_pstrdup(p, apr_dbd_get_entry(dbd->driver, row, 3));
        }
    }
    if (back_link && *back_link)
    {
        apr_table_set(r->subprocess_env, "psiphon_custom_user_agent_back_link", back_link);
    }
    return user_agent_string;
}


static int psiphon_headers_fixup(request_rec *r)
{
    psiphon_headers_ctxt* ctx;
    psiphon_headers_conf* conf;
    const char* user_agent_string = 0;

    conf =(psiphon_headers_conf *) ap_get_module_config(r->per_dir_config,
        &psiphon_headers_module);

    /* Say no to subrequests */
    if(r->main != NULL)
    {
        return DECLINED;
    }

    /* Not a reversed proxy request*/
    if(!r->proxyreq || r->proxyreq != PROXYREQ_REVERSE || 
        !r->filename || strncmp(r->filename, "proxy:", 6) != 0)
    {
        return DECLINED;
    }

    /*module_disabled*/
    if(!conf->psiphon_headers_enable)
    {
        return DECLINED;
    };

    ctx = create_context(r);

    if(!ctx)
    {
        //skip processing if no context created
        return DECLINED;
    }

    /********************************************************************
    INPUT HEADERS
    /********************************************************************/
    /*  Anonymize request unset "Via" and X-Forwarded*/
    apr_table_unset(r->headers_in, "Via");
    apr_table_unset(r->headers_in, "X-Forwarded-For");

    /* We don't want any server related cookies to be passed to the remote site -
       unset em */
    apr_table_unset(r->headers_in, "Cookie");

    user_agent_string = get_user_agent_for_this_request(r);
    if (user_agent_string && *user_agent_string)
    {
        /* If we found a non 0 length user agent string, then set the custom user agent */
        apr_table_set(r->headers_in, "User-Agent", user_agent_string);
        apr_table_set(r->subprocess_env, "psiphon_custom_user_agent", user_agent_string);
    }

    /* Add cookies headers from DB for authenticated user*/
    if(!psiphon_db_to_cookie(r, ctx))
    {
        return HTTP_INTERNAL_SERVER_ERROR;
    }

    /*set correct Referer header*/
    fixup_referer(r, ctx);

    /*Log headers*/
    apr_table_do(log_header, "input", r->headers_in, NULL);

    /********************************************************************
    OUTPUT HEADERS
    /********************************************************************/
    ap_add_output_filter(PSIPHON_HEADERS_FNAME, ctx, r, r->connection);

    return OK;
}


static void* psiphon_headers_dir_config(apr_pool_t* pool, char* dir)
{
    psiphon_headers_conf* ret = apr_pcalloc(pool, sizeof(psiphon_headers_conf)) ;
    ret->psiphon_headers_enable = 0;
    ret->loc = apr_pstrdup(pool, dir);

    return ret ;
}


static const char* set_psiphon_headers(cmd_parms *parms, void *in_dconf, int arg)
{
    psiphon_headers_conf *dconf = in_dconf;
    dconf->psiphon_headers_enable = arg;
    if(!dconf->loc)
    {
        return "psiphon_headers_module must be set in per-directory context only";
    }
    if(dconf->psiphon_headers_enable)
    {
        /* prepare statements for psiphon_cookie_to_db() */
        ap_dbd_prepare(parms->server,
            "delete from cookie where user=%d and domain=%s and path=%s and name=%s",
                "psiphon_delete_cookie");
        ap_dbd_prepare(parms->server,
            "insert into cookie(user, domain, path, name, value, expire, secure) "
            "values(%d, %s, %s, %s, %s, %d, %d) "
            "ON DUPLICATE KEY update "
            "value=VALUES(value), expire=VALUES(expire), secure=VALUES(secure)",
                "psiphon_insert_cookie");

        /* prepare select statement for psiphon_db_to_cookie() */
        ap_dbd_prepare(parms->server,
            "select name, path, concat(name,\"=\",value) as v from cookie "
            "where user=%d and domain=%s and (expire>%d or expire=-1) "
            "and secure=0 order by id desc",
                "psiphon_select_http_cookie");
        ap_dbd_prepare(parms->server,
            "select name, path, concat(name,\"=\",value) as v from cookie "
            "where user=%d and domain=%s and (expire>%d or expire=-1) "
            "order by id desc",
                "psiphon_select_https_cookie");
    }
    return NULL;
}


static const command_rec psiphon_dir_cmds[] =
{
    AP_INIT_FLAG("PsiphonHeadersFixup", set_psiphon_headers, NULL, OR_ALL,
    "on/off flag"),
    {NULL}
};

static void psiphon_headers_hooks(apr_pool_t* p)
{
    ap_register_output_filter(PSIPHON_HEADERS_FNAME, output_headers_filter,
        NULL, AP_FTYPE_RESOURCE) ;
    ap_hook_fixups(psiphon_headers_fixup, NULL, NULL, APR_HOOK_FIRST);
}


module AP_MODULE_DECLARE_DATA psiphon_headers_module =
{
    STANDARD20_MODULE_STUFF,
    psiphon_headers_dir_config,
    NULL,
    NULL,
    NULL,
    psiphon_dir_cmds,
    psiphon_headers_hooks
} ;
