/*
    Psiphon Circumvention Platform
    Copyright (C) 2009 Psiphon Inc.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_protocol.h"
#include "http_core.h"
#include "http_main.h"
#include "http_log.h"
#include "ap_mpm.h"
#include "apr_strings.h"
#include "scoreboard.h"


#define HAVE_SHMGET

#ifdef HAVE_SHMGET
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#endif


#define DEBUG
#undef DEBUG


#ifdef DEBUG
#define lsconn_log(args...) syslog(LOG_DEBUG, args)
#else
#define lsconn_log(args...)  ""
#endif



module AP_MODULE_DECLARE_DATA limit_sessconn_module;

#if (defined(__GNU_LIBRARY__) && !defined(_SEM_SEMUN_UNDEFINED)) || defined(__FreeBSD__)
/* union semun is defined by including <sys/sem.h> */
#else
/* according to X/OPEN we have to define it ourselves */
union semun 
{
	int val;
	struct semid_ds *buf;
	unsigned short *array;
	struct seminfo *__buf;
};
#endif

typedef struct 
{
	int uid;
	pid_t pid;
	int lock;
} limit_sessconn_score_rec;

typedef struct 
{
	int enabled;
	int max_sessions;
} limit_sessconn_config;

typedef struct 
{
	int shmid;
	int semid;
	int server_limit;
	int thread_limit;
	int max_clients;
	pid_t child_pid;
	limit_sessconn_score_rec* score_board;
} limit_sessconn_ctx;

static  limit_sessconn_ctx* ctxt;

/******************Helper funcs**********************************/
int lmt_is_number(const char *str)
{
	if (!str)
		return(1);
	int len = strlen(str);
	int i;

	for (i = 0; i < len; i++) 
	{
		if (!isdigit(str[i])) 
		{
			return(0);
		}
	}
	return(1);
}

/*****************************************************************/
#ifdef HAVE_SHMGET
static apr_status_t limit_sessconn_detach_shm(void *data)
{
	return shmdt(data);
}
#endif

static apr_status_t limit_sessconn_sem_remove(void* dummy)
{
	union semun arg;

	arg.val   = 0;
	return semctl(ctxt->semid, 0, IPC_RMID, arg);
}

void limit_sessconn_lock(int semid)
{
	struct sembuf sops;

	sops.sem_num  = 0;
	sops.sem_op   = -1;
	sops.sem_flg  = SEM_UNDO;

	semop(semid, &sops, 1);
}

void limit_sessconn_unlock(int semid)
{
	struct sembuf sops;

	sops.sem_num  = 0;
	sops.sem_op   = 1;
	sops.sem_flg  = SEM_UNDO;

	semop(semid, &sops, 1);
}

static limit_sessconn_score_rec* score_board_cleanup()
{

	worker_score *ws_record;
	process_score *ps_record;

	limit_sessconn_score_rec* rec;
	limit_sessconn_score_rec* last_empty_rec = NULL;
	int i, j, k;
	int rec_match;

	for (i = 0; i < ctxt->max_clients; i++) 
	{
		rec =  &ctxt->score_board[i];
		rec_match = 0;
		if (!rec->pid || !rec->uid) 
		{
			last_empty_rec = rec;
			continue;
		}

		for (j = 0; j < ctxt->server_limit; j++) 
		{
			ps_record = ap_get_scoreboard_process(j);
			if (rec->pid == ps_record->pid) 
			{

				for (k = 0; k < ctxt->thread_limit; k++) 
				{
					ws_record = ap_get_scoreboard_worker(j, k);
					switch (ws_record->status) 
					{
						case SERVER_BUSY_READ:
						case SERVER_BUSY_WRITE:
						case SERVER_BUSY_KEEPALIVE:
							rec_match = 1;
							break;
					}
					if (rec_match) break;
				}
			}
			if (rec_match) break;

		}
		if (rec_match) continue;

		rec->pid = 0;
		rec->uid = 0;
		last_empty_rec = rec;
	}

	return last_empty_rec;
}

static int check_limit(int user_id, int max_sessions)
{
	int i;
	limit_sessconn_score_rec* rec;
	int sessions_count = 0;
	for (i = 0; i < ctxt->max_clients; i++) 
	{
		rec =  &ctxt->score_board[i];
		if (rec->uid == user_id) 
		{
			sessions_count++;
		}
	}
	lsconn_log("sessions_count = %d", sessions_count);
	return sessions_count > max_sessions;
}

static int update_score_board(limit_sessconn_score_rec* rec, int uid)
{
	if(rec->pid || rec->uid)
	{
		return 1;
	}
	rec->pid = ctxt->child_pid;
	rec->uid = uid;
	return 0;
}

static void *create_dir_config(apr_pool_t *p, char *d)
{
	limit_sessconn_config *dcfg = (limit_sessconn_config *)
	                              apr_pcalloc(p, sizeof (*dcfg));
	return (void*)dcfg;
}

static void *merge_dir_config(apr_pool_t* pool, void* BASE, void* ADD)
{
	limit_sessconn_config* base = (limit_sessconn_config*) BASE ;
	limit_sessconn_config* add = (limit_sessconn_config*) ADD ;
	limit_sessconn_config* conf = apr_palloc(pool, sizeof(limit_sessconn_config)) ;

	conf->enabled = (add->enabled == 0) ? base->enabled : add->enabled;
	conf->max_sessions = (add->max_sessions == 0) ? base->max_sessions : add->max_sessions;

	return conf;
}

static const char *conn_per_sess_cmd(cmd_parms *parms, void *CFG, const char *arg)
{
	limit_sessconn_config *cfg = CFG;

	if (lmt_is_number(arg)) {
		cfg->max_sessions = atoi(arg);
	} else {
		cfg->max_sessions = 0;
		return "[mod_limitsessconn] MaxConnectionsPerSession is not a valid number";
	}
	if ((cfg->max_sessions > 256) || (cfg->max_sessions < 0)) {
		return "[mod_limitsessconn] MaxConnectionsPerSession must be between 0 and 256";
	}

	return(NULL);
}

static const char *limit_sessconn_enable(cmd_parms *parms, void *in_dconf, int arg)
{
	limit_sessconn_config *dconf = in_dconf;
	dconf->enabled = arg;
	return NULL;
}

void limit_sessconn_sem_init(int semid)
{
	union semun arg;
	unsigned short values[1];

	values[0] = 1;
	arg.val   = 1;
	arg.array = values;

	semctl(semid, 0, SETALL, arg);
}

static void pid_child_init(apr_pool_t *p, server_rec *s)
{
	ctxt->child_pid = getpid();
}

static int limit_sessconn_init(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s)
{
	size_t score_board_size;
	int i;

	ctxt = apr_palloc(p,sizeof(limit_sessconn_ctx));

	ap_mpm_query(AP_MPMQ_HARD_LIMIT_THREADS, &ctxt->thread_limit);
	ap_mpm_query(AP_MPMQ_HARD_LIMIT_DAEMONS, &ctxt->server_limit);
	ap_mpm_query(AP_MPMQ_MAX_DAEMON_USED, &ctxt->max_clients);

	score_board_size = sizeof(limit_sessconn_score_rec)
	                   * (ctxt->max_clients);

	ctxt->score_board = apr_palloc(p, score_board_size);
	memset(ctxt->score_board, 0, score_board_size);

#ifdef HAVE_SHMGET
	if ((ctxt->shmid = shmget(IPC_PRIVATE, score_board_size, IPC_CREAT | 0662)) == -1) 
	{
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL,
		             "[mod_limitsessconn] could not initialize shared memory");
		exit(1);
	}


	if ( (ctxt->score_board = (limit_sessconn_score_rec *) shmat(ctxt->shmid, NULL, 0)) ==  (limit_sessconn_score_rec *)-1) 
	{
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL,
		             "[mod_limitsessconn] ] shmat error");
		/* exit later after marking the segment to remove itself */
	} 
	else 
	{
		apr_pool_cleanup_register(p, (void *)ctxt->score_board, limit_sessconn_detach_shm, apr_pool_cleanup_null);
	}

	/* Mark the segment for deletion once all attachments are detached */
	if (shmctl(ctxt->shmid, IPC_RMID, NULL) == -1) 
	{
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL,
		             "[mod_limitsessconn] Could not mark shared segment for deletion, you must manually clean it up");
	}

	/* exit if we didn't attach successfully */
	if (ctxt->score_board == (limit_sessconn_score_rec *) -1) 
	{
		exit(1);
	}

#if 0
	/* setup semaphore */
	
	if (ctxt->semid = semget(IPC_PRIVATE, 1, IPC_CREAT | 0666) == -1) 
	{
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL,
		             "[mod_limitsessconn] semget error");
		exit(1);
	} 
	else 
	{
		/*register semaphore for cleanup*/
		apr_pool_cleanup_register(p, NULL, limit_sessconn_sem_remove, apr_pool_cleanup_null);
	}
	limit_sessconn_sem_init(ctxt->semid);
#endif	
#else
	ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL,
	             "[mod_limitsessconn] Shared memory is not available, "
	             "MaxConnectionsPerSession will have no effect");
#endif



	return OK;
}

static int limit_sessconn_handler(request_rec *r)
{
	int i, j;
	worker_score *ws_record;
	const char* uid_str;
	int uid;
	limit_sessconn_score_rec* current_rec;
	apr_time_t tstart, tend;

	limit_sessconn_config *dconf =
	    ap_get_module_config(r->per_dir_config, &limit_sessconn_module);

	if (!ap_is_initial_req(r) || !dconf->enabled) 
	{
		return DECLINED;
	}

	if (!(uid_str = apr_table_get(r->subprocess_env, "psiphon_uid"))) 
	{
		return DECLINED;
	}

	if (!(uid = atoi(uid_str))) 
	{
		return DECLINED;
	}

	tstart = tend = apr_time_now();
	current_rec = score_board_cleanup();
	tend = apr_time_now();
	lsconn_log("[mod_limitsessconn] score_board_cleanup() time = %d", (int)(tend-tstart));

	if (check_limit(uid, dconf->max_sessions)) 
	{
		return HTTP_SERVICE_UNAVAILABLE;
	}

	if (current_rec && uid)
	{
		if(update_score_board(current_rec, uid))
		{
			ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
				"mod_limisessconn: 'rec' already changed by another process [racing condition]");
			lsconn_log("mod_limisessconn: 'rec' already changed by another process [racing condition]");
		}
	}

	return DECLINED;
}

static command_rec cmds[] = 
{
	AP_INIT_FLAG("LimitSessionConn", limit_sessconn_enable, NULL, OR_ALL,
	"on/off flag"),
	AP_INIT_TAKE1("MaxConnPerSession", conn_per_sess_cmd, NULL, OR_LIMIT|RSRC_CONF,
	"maximum simultaneous connections per IP address"),
	{NULL},
};

static void register_hooks(apr_pool_t *p)
{
	static const char * const before_me[] = { "mod_psiphon_auth.c", NULL };
	ap_hook_child_init(pid_child_init, NULL, NULL, APR_HOOK_MIDDLE);
	ap_hook_post_config(limit_sessconn_init, NULL, NULL, APR_HOOK_MIDDLE);
	ap_hook_access_checker(limit_sessconn_handler, before_me, NULL, APR_HOOK_LAST);
}

module AP_MODULE_DECLARE_DATA limit_sessconn_module = 
{
	STANDARD20_MODULE_STUFF,
	create_dir_config,
	merge_dir_config,
	NULL,
	NULL,
	cmds,
	register_hooks
};
