/*
    Psiphon Circumvention Platform
    Copyright (C) 2009 Psiphon Inc.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
#Make sure that mod_proxy is fully loaded, otherwise it will return HTTP_FORBIDDEN

LoadModule map_to_proxy_module  modules/mod_map_to_proxy.so
....
......
<Location /fetch>
MapToProxy on|off
.........
*/

#include "apr_strings.h"
#include "apr_hash.h"

#define APR_WANT_STRFUNC
#define APR_WANT_IOVEC
#include "apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_core.h"
#include "http_protocol.h"
#include "psiphon_url.h"
#include "mod_proxy.h"

#define DEFAULT_MAP2PROXY  0

#define DEBUG
#undef DEBUG

#ifdef DEBUG
#define map2proxy_log(FMT, ARG...) syslog(LOG_DEBUG, "%s:%s:%d: " FMT, __FILE__, __FUNCTION__, __LINE__, ## ARG)
#else
#define map2proxy_log(args...)  ""
#endif

#define PSIPHON_SETUP_TRANSFORM_FNAME "set-mod-transform"

struct ap_filter_provider_t
{
    /** How to match this provider to filter dispatch criterion */
    enum
    {
        STRING_MATCH,
        STRING_CONTAINS,
        REGEX_MATCH,
        INT_EQ,
        INT_LT,
        INT_LE,
        INT_GT,
        INT_GE,
        DEFINED
    } match_type;

    /** negation on match_type */
    int not;

    /** The dispatch match itself - union member depends on match_type */
    union
    {
        const char *string;
        ap_regex_t *regex;
        int         number;
    } match;

    /** The filter that implements this provider */
    ap_filter_rec_t *frec;

    /** The next provider in the list */
    ap_filter_provider_t *next;

    /** Dispatch criteria for filter providers */
    enum
    {
        HANDLER,
        REQUEST_HEADERS,
        RESPONSE_HEADERS,
        SUBPROCESS_ENV,
        CONTENT_TYPE
    } dispatch;

    /** Match value for filter providers */
    const char* value;
};

module AP_MODULE_DECLARE_DATA map_to_proxy_module;

static int proxy_available;

typedef struct 
{
    int map2proxy;
    int url_encoded;
    const char* loc;
} map_to_proxy_dir_config;

typedef struct 
{
    apr_array_header_t *proxyblock;
} map_to_proxy_server_config;

static int is_base64_legal(char *str)
{
    /**
       to satisfy the condition the input str must be of length multiple of 4
       and every character of the string must be [A-Za-z0-9/+=]
     */

    char *c = 0;
    int is_base64_char;
    if(!str || !*str)
    {
        return 0;
    }
    if(strlen(str)%4 != 0)
    {
        return 0;
    }

    for(c = str; *c; c++)
    {
        is_base64_char = (*c >= 'A' && *c <= 'Z') || (*c >= 'a' && *c <= 'z') ||
            (*c >= '0' && *c <= '9') || (*c == '+') ||
            (*c == '/') || (*c == '=');

        if(!is_base64_char)
        {
            return 0;
        }
    }
    return 1;
}

static void *create_server_config(apr_pool_t * p, server_rec * x)
{
    map_to_proxy_server_config *cfg = apr_pcalloc(p, sizeof(map_to_proxy_server_config));
    cfg->proxyblock = apr_array_make(p, 10, sizeof(struct dirconn_entry));
    return cfg;
}

static void *create_dir_config(apr_pool_t *p, char *dir)
{
    map_to_proxy_dir_config *new =
        (map_to_proxy_dir_config *) apr_pcalloc(p, sizeof(map_to_proxy_dir_config));

    new->map2proxy = DEFAULT_MAP2PROXY;
    new->loc = apr_pstrdup(p, dir);

    return (void *) new;
}

//Based on set_proxy_dirconn from mod_proxy
static const char *set_block_map_to_proxy(cmd_parms *parms, void *dummy, const char* arg)
{
    server_rec *s = parms->server;
    map_to_proxy_server_config *conf =
            ap_get_module_config(s->module_config, &map_to_proxy_module);

    struct dirconn_entry *New;
    struct dirconn_entry *list = (struct dirconn_entry *) conf->proxyblock->elts;
    int found = 0;
    int i;

    /* Don't duplicate entries */
    for (i = 0; i < conf->proxyblock->nelts; i++)
    {
        if (strcasecmp(arg, list[i].name) == 0)
            found = 1;
    }

    if (!found)
    {
        New = apr_array_push(conf->proxyblock);
        New->name = apr_pstrdup(parms->pool, arg);
        New->hostaddr = NULL;
    }
    if (ap_proxy_is_ipaddr(New, parms->pool))
    {
        map2proxy_log("parsed addr %s", inet_ntoa(New->addr));
        map2proxy_log("parsed mask %s", inet_ntoa(New->mask));
    }
    else if (ap_proxy_is_domainname(New, parms->pool))
    {
        ap_str_tolower(New->name);
        map2proxy_log("parsed domain %s", New->name);
    }
    else if (ap_proxy_is_hostname(New, parms->pool))
    {
        ap_str_tolower(New->name);
        map2proxy_log("parsed host %s", New->name);
    }
    else
    {
        map2proxy_log("parsed word %s", New->name);
    }
    return NULL;
}

static const char *set_url_encoded(cmd_parms *parms, void *in_dconf, int arg)
{
    map_to_proxy_dir_config *dconf = in_dconf;
    dconf->url_encoded = arg;
    return NULL;
}

static const char *set_map_to_proxy(cmd_parms *parms, void *in_dconf, int arg)
{
    map_to_proxy_dir_config *dconf = in_dconf;
    dconf->map2proxy = arg;
    if(!dconf->loc)
    {
        return "map_to_proxy_module must be set in per-directory context only";
    }
    return NULL;
}

static apr_status_t setup_mod_transform_filter(ap_filter_t* f, apr_bucket_brigade* bb)
{
    ap_filter_provider_t *provider;
    ap_filter_rec_t *filter;
    ap_filter_t* ff = f;
    ap_filter_t* rm_ff = NULL;
    const char* remove_transform_filter;

    struct dirconn_entry* entry;

    remove_transform_filter = apr_table_get(f->r->subprocess_env, "psiphon_remove_mod_transform");

    while (ff && ff->frec)
    {
        rm_ff = NULL;
        filter = ff->frec;
        for (provider = filter->providers; provider; provider = provider->next)
        {
            if(provider->frec->name && strncmp("xslt", provider->frec->name, 4) == 0)
            {
                if(remove_transform_filter && *remove_transform_filter)
                    rm_ff = ff;
            }
        }
        ff = ff->next;
        if(rm_ff)
        {
            ap_remove_output_filter(rm_ff);
        }
    }
    ap_remove_output_filter(f);
    return ap_pass_brigade(f->next,bb);
}

static int hook_uri2proxy(request_rec *r)
{
    int rv;
    map_to_proxy_dir_config *dconf;
    map_to_proxy_server_config *sconf;
    char *proxy_uri = NULL;
    char *return_url = NULL;
    char *uri = NULL;
    apr_uri_t parsed_proxy_uri;
    char* site;
    char* proxy_fname;
    char* proxy_args;
    
    const char* tmp_hostname;
    char* tmp_uri;
    
    int i;
    int scheme_found = 0;
    int block_connect = 0;
    struct dirconn_entry *list;

    apr_array_header_t *schemes;
    apr_pool_t *tmp_pool;
    apr_pool_t *p = r->pool;

    map2proxy_log("r->proxyreq == %s",  r->proxyreq == PROXYREQ_REVERSE?"PROXYREQ_REVERSE":"PROXYREQ_NONE");
    if(r->proxyreq == PROXYREQ_REVERSE)
    {
        dconf = (map_to_proxy_dir_config *) ap_get_module_config(r->per_dir_config,
                &map_to_proxy_module);

        sconf = (map_to_proxy_server_config *)
            ap_get_module_config(r->server->module_config, &map_to_proxy_module);


        map2proxy_log("dconf->map2proxy: %d, dconf->loc: %s", dconf->map2proxy , dconf->loc);
        if (!dconf->map2proxy || !dconf->loc)
        {
            return DECLINED;
        };

        proxy_available = (ap_find_linked_module("mod_proxy.c") != NULL);
        if (!proxy_available)
        {
            return HTTP_FORBIDDEN;
        }

        if (r->uri == NULL)
        {
            return HTTP_INTERNAL_SERVER_ERROR;
        }

        r->uri = r->unparsed_uri;

        if(dconf->url_encoded && is_base64_legal(&r->unparsed_uri[strlen(dconf->loc) + 1]))
        {
            //uri = psi_rawurl_decode(&r->unparsed_uri[strlen(dconf->loc) + 1]);
            //Using base64 instead of raw url encoding due to problems with JW FLV player
            uri = ap_pbase64decode(r->pool, &r->unparsed_uri[strlen(dconf->loc) + 1]);
        }
        else
        {
            uri = &r->unparsed_uri[strlen(dconf->loc) + 1];
        }

        map2proxy_log("got uri: %s", uri);

        if(!uri || !*uri)
        {
            map2proxy_log("no uri");
            return DECLINED;
        }
        /*check if scheme matches one of 'ftp' 'http' or 'https'*/
        if (apr_pool_create(&tmp_pool, r->pool) != APR_SUCCESS)
        {
            return HTTP_INTERNAL_SERVER_ERROR;
        }

        schemes = apr_array_make(tmp_pool, 3, sizeof(const char *));

        *(const char **)apr_array_push(schemes) = "ftp://";
        *(const char **)apr_array_push(schemes) = "http://";
        *(const char **)apr_array_push(schemes) = "https://";

        char **scheme = (char **) schemes->elts;

        for (i = 0; i < schemes->nelts; ++i) 
        {
            if(scheme[i] && !strncmp(uri, scheme[i], strlen(scheme[i])))
            {
                map2proxy_log("scheme %s found in %s", scheme[i], uri);
                scheme_found = 1;
            }
        }
        apr_pool_destroy(tmp_pool);
        if(!scheme_found)
        {
            map2proxy_log("no matching scheme found in uri: %s", uri);
            /*Revert to non-proxy*/
            r->proxyreq = PROXYREQ_NONE;
            return DECLINED;
        }
        /*end check scheme*/
        apr_table_setn(r->subprocess_env, "psiphon_proxy_uri", uri);

        rv = apr_uri_parse(p, uri, &parsed_proxy_uri);

        if (rv != APR_SUCCESS)
        {
            return HTTP_BAD_REQUEST;
        }
        if (!parsed_proxy_uri.hostname || !parsed_proxy_uri.scheme)
        {
            return HTTP_BAD_REQUEST;
        }

        site = apr_uri_unparse(p, &parsed_proxy_uri, APR_URI_UNP_OMITPASSWORD | APR_URI_UNP_OMITPATHINFO);
        apr_table_set(r->subprocess_env, "psiphon_proxy_site", site);

        //store uri and hostname
        tmp_uri = r->uri;
        tmp_hostname = r->hostname;

        //Check if request should be blocked        
        r->uri = apr_pstrdup(p, uri);
        r->hostname = parsed_proxy_uri.hostname;

        list = (struct dirconn_entry *)sconf->proxyblock->elts;

        for (block_connect = i = 0; i < sconf->proxyblock->nelts &&
                !block_connect; i++) 
        {
            block_connect = list[i].matcher(&list[i], r);
        }

        if(block_connect)
        {
            map2proxy_log("connect blocked by BlockMapToProxy for %s", r->hostname);
            return HTTP_FORBIDDEN;
        }

        //restore vars
        r->uri = tmp_uri;
        r->hostname = tmp_hostname;

        r->filename = apr_psprintf(p, "proxy:%s", uri);

        apr_table_setn(r->notes, "proxy-nocanon", "1");

        ap_add_output_filter(PSIPHON_SETUP_TRANSFORM_FNAME, NULL, r, r->connection);

        return OK;
    }
    return DECLINED;
}

static int set_proxy_request(request_rec *r)
{
    /*
     *  check if request uri starts with one of the proxy prefixes 
     */
    apr_array_header_t *prefixes;
    apr_array_header_t *schemes;
    apr_pool_t *tmp_pool;
    int i, j;
    char* uri = r->unparsed_uri;


    if (apr_pool_create(&tmp_pool, r->pool) != APR_SUCCESS)
    {
        return HTTP_INTERNAL_SERVER_ERROR;

    }
    

    prefixes = apr_array_make(tmp_pool, 4, sizeof(const char *));

    *(const char **)apr_array_push(prefixes) = apr_table_get(r->subprocess_env, "psiphon_html_prefix");
    *(const char **)apr_array_push(prefixes) = apr_table_get(r->subprocess_env, "psiphon_js_prefix");
    *(const char **)apr_array_push(prefixes) = apr_table_get(r->subprocess_env, "psiphon_css_prefix");
    *(const char **)apr_array_push(prefixes) = apr_table_get(r->subprocess_env, "psiphon_flash_prefix");

    schemes = apr_array_make(tmp_pool, 3, sizeof(const char *));

    *(const char **)apr_array_push(schemes) = "ftp://";
    *(const char **)apr_array_push(schemes) = "http://";
    *(const char **)apr_array_push(schemes) = "https://";

    char **prefix = (char **) prefixes->elts;
    char **scheme = (char **) schemes->elts;

    for (i = 0; i < prefixes->nelts; ++i) 
    {
        if (uri && prefix[i] && !strncmp(uri, prefix[i], strlen(prefix[i])))
        {
            uri += strlen(prefix[i]);
            if(uri[0] != '/')
                continue;
            uri++;
            if(uri[0] == '\0')
                continue;

            map2proxy_log("prefix: %s found in %s", prefix[i], uri);
            r->proxyreq = PROXYREQ_REVERSE;
            break;
        }
    }
    apr_pool_destroy(tmp_pool);
    return OK;
}

static const command_rec dir_cmds[] =
{
    AP_INIT_FLAG("MapToProxy", set_map_to_proxy, NULL, ACCESS_CONF,
            "on/off flag"),
    AP_INIT_FLAG("MapURLEncoded", set_url_encoded, NULL, ACCESS_CONF,
            "on/off flag"),
    AP_INIT_ITERATE("BlockMapToProxy", set_block_map_to_proxy, NULL, RSRC_CONF,
            "A list of domains, hosts, or subnets to block"),
    {NULL}
};

static void register_hooks(apr_pool_t *p)
{
    static const char * const aszPre[]={ "mod_dir.c", "mod_psiphon_auth.c", NULL };

    ap_hook_post_read_request(set_proxy_request, NULL, NULL, APR_HOOK_LAST);
    ap_hook_translate_name(hook_uri2proxy, aszPre, NULL, APR_HOOK_REALLY_FIRST);
    ap_register_output_filter(PSIPHON_SETUP_TRANSFORM_FNAME, setup_mod_transform_filter,
        NULL, AP_FTYPE_RESOURCE) ;
}


/* the main config structure */
module AP_MODULE_DECLARE_DATA map_to_proxy_module =
{
    STANDARD20_MODULE_STUFF,
    create_dir_config,          /* create per-dir    config structures */
    NULL,                        /* merge  per-dir    config structures */
    create_server_config,        /* create per-server config structures */
    NULL,                        /* merge  per-server config structures */
    dir_cmds,                    /* table of config file commands       */
    register_hooks               /* register hooks                      */
};

/*EOF*/
