/*
    Psiphon Circumvention Platform
    Copyright (C) 2009 Psiphon Inc.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#define APR_WANT_STRFUNC
#define APR_WANT_IOVEC
#include "apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#include <apr_strings.h>
#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"

#include "apr_optional.h"
#include "mod_post2get.h"


#define DEFAULT_POST2GET  0


#define DEBUG
#undef DEBUG


#ifdef DEBUG
#define post2get_log(args...) syslog(LOG_DEBUG, args)
#else
#define post2get_log(args...)  ""
#endif


module AP_MODULE_DECLARE_DATA post2get_module;

static int proxy_available;

typedef struct {
    int post2get;
    const char* loc;
} post2get_dir_config;

typedef struct {
    const char* module_location;
} post2get_srv_config;

static void *create_post2get_dirconfig(apr_pool_t *p, char *dir)
{
    post2get_dir_config *new =
            (post2get_dir_config *) apr_pcalloc(p, sizeof(post2get_dir_config));

    new->post2get = DEFAULT_POST2GET;
    new->loc = apr_pstrdup(p, dir);
    return (void *) new;
}
static void *create_post2get_srvconfig(apr_pool_t * p, server_rec * x)
{
    post2get_srv_config *cfg = apr_pcalloc(p, sizeof(post2get_srv_config));
    cfg->module_location = NULL;
    return cfg;
}
static int register_module_location(apr_pool_t* pool, server_rec* s, const char* location)
{
    post2get_srv_config *conf =
            ap_get_module_config(s->module_config, &post2get_module);
    if(conf->module_location == NULL)
    {
        conf->module_location = apr_pstrdup(pool, location);
        return 0;
    }
    return 1;
}

static const char *set_post2get(cmd_parms *parms, void *in_dconf, int arg)
{
    post2get_dir_config *dconf = in_dconf;
    dconf->post2get = arg;
    if(!dconf->loc)
    {
        return "post2get_module must be set in per-directory context only";
    }
    if(register_module_location(parms->pool, parms->server, dconf->loc) !=0 )
    {
        return "post2get_module must be set only once per server, no multiple locations allowed";
    }
    return NULL;
}

static int post2get_handler(request_rec *r)
{
    apr_array_header_t *a;
    apr_pool_t *p = r->pool;
    apr_pool_t *temp_pool;
    int res, ret, len = 0;
    char* query = NULL;
    char* path = NULL;
    char* qmark = NULL;
    char* redirect_uri = NULL;
    char* result_uri = NULL;
    const char* html_prefix = apr_table_get(r->subprocess_env, "psiphon_html_prefix");
    

    if (r->main != NULL)  /*no subrequests*/
    {
        return DECLINED;
    }

    
    if(!html_prefix)
    {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "html prefix is empty.") ;
        return HTTP_INTERNAL_SERVER_ERROR;
    }

    
    void *dconf = r->per_dir_config;
    post2get_dir_config *dirconf =
            (post2get_dir_config *) ap_get_module_config(dconf, &post2get_module);


    if(!dirconf->post2get)
        return DECLINED;
    
    if(!dirconf->loc)
        return HTTP_INTERNAL_SERVER_ERROR;

    path = &r->uri[strlen(dirconf->loc) + 1];

    if(!path)
        return DECLINED;

    if ((r->uri[0] != '/' && r->uri[0] != '\0') || !dirconf->post2get) {
        return DECLINED;
    }
    if (r->method_number!=M_POST)
        return DECLINED;


    if (apr_pool_create(&temp_pool, p) != APR_SUCCESS)
        return DECLINED;
    
    a = apr_array_make(temp_pool, 1, sizeof(char *));
    
    ret = ap_setup_client_block(r, REQUEST_CHUNKED_ERROR);

    if (ret != OK) {
        return HTTP_INTERNAL_SERVER_ERROR;
    }
    
    char temp_buf[HUGE_STRING_LEN];
    if (ap_should_client_block(r)) {
        
        while ((res = ap_get_client_block(r, temp_buf, HUGE_STRING_LEN)) > 0)
        {
            len+=res;
            if(len>HUGE_STRING_LEN)
                break;
            *(temp_buf + res) = '\0';
            *(const char **)apr_array_push(a) = apr_pstrdup(temp_pool, temp_buf);
         }
    } else {
        //hits this condition when post body length == 0
        *(const char **)apr_array_push(a) = '\0';
    }

    query = apr_array_pstrcat(p, a, '\0');
    query = apr_pstrcat(p, "?", query, '\0');

    qmark = strchr(path, '?');
    if (qmark != NULL)
        *qmark = '\0';

    redirect_uri = apr_pstrcat(p, path, query, NULL);

    result_uri = apr_pstrcat(p,  html_prefix,
            "/",
            (const char*)redirect_uri,
            NULL);
    
    apr_table_setn(r->headers_out, "Location", result_uri);
    return HTTP_MOVED_PERMANENTLY;

    
}

const char* psiphon_get_post2get_prefix(request_rec *r)
{
    post2get_srv_config *sconf;
    sconf = (post2get_srv_config *)
            ap_get_module_config(r->server->module_config, &post2get_module);
    return sconf->module_location;
}
static const command_rec dir_cmds[] =
{
    AP_INIT_FLAG("PsiphonPost2Get", set_post2get, NULL, ACCESS_CONF,
                 "on/off flag"),
                 {NULL}
};

static void register_hooks(apr_pool_t *p)
{
    ap_hook_handler(post2get_handler,NULL,NULL,APR_HOOK_MIDDLE);
    APR_REGISTER_OPTIONAL_FN(psiphon_get_post2get_prefix);

}

    /* the main config structure */
module AP_MODULE_DECLARE_DATA post2get_module = {
    STANDARD20_MODULE_STUFF,
    create_post2get_dirconfig, /* create per-dir    config structures */
    NULL,                       /* merge  per-dir    config structures */
    create_post2get_srvconfig,  /* create per-server config structures */
    NULL,                       /* merge  per-server config structures */
    dir_cmds,                   /* table of config file commands       */
    register_hooks              /* register hooks                      */
};

