/*
    Psiphon Circumvention Platform
    Copyright (C) 2009 Psiphon Inc.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/********************************************************************
  Copyright (c) 2005-6, WebThing Ltd
  Author: Nick Kew <nick@webthing.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*********************************************************************/

#include <ctype.h>
#include <syslog.h>

#include <http_protocol.h>
#include <http_config.h>
#include <http_log.h>
#include <apr_strings.h>
#include <apr_hash.h>
#include <apr_uri.h>
#include <apr_strmatch.h>

#include <apr_optional.h>

#include "mod_proxy_css.h"
#include "psiphon_url.h"
#include "css_scanner.h"
#include "stringbuf.h"

#define DEBUG
#undef DEBUG

#ifdef DEBUG
#define proxy_css_log(args...) syslog(LOG_DEBUG, args)
#else
#define proxy_css_log(args...)  ""
#endif


typedef struct
{
    enum
    {
        LINEEND_UNSET,
        LINEEND_ANY,
        LINEEND_UNIX,
        LINEEND_MAC,
        LINEEND_DOS,
        LINEEND_CUSTOM,
        LINEEND_NONE
    } lineend ;
    int proxy_css;
    int lechar;
    const char* loc;
} proxy_css_dir_conf ;

typedef struct
{
    const char* module_location;
} proxy_css_srv_conf;

module AP_MODULE_DECLARE_DATA proxy_css_module ;

typedef struct
{
    apr_bucket_brigade* bbsave ;
    const char* htmlprefix ;
    const char* cssprefix ;
} proxy_css_ctx ;

const char* psiphon_get_css_prefix(request_rec *r)
{
    proxy_css_srv_conf *sconf;
    sconf = (proxy_css_srv_conf *)
        ap_get_module_config(r->server->module_config, &proxy_css_module);
    return sconf->module_location;
}

static apr_status_t proxy_css_filter(ap_filter_t* f, apr_bucket_brigade* bb)
{
    int i, j;
    int found;
    const char* bufp;
    apr_size_t bytes ;
    apr_size_t fbytes ;
    apr_size_t offs ;
    const char* buf ;
    const char* le = NULL ;
    const char* le_n ;
    const char* le_r ;
    char* fbuf ;
    apr_bucket* b = APR_BRIGADE_FIRST(bb) ;
    apr_bucket* b1 ;
    apr_status_t rv ;
    char* proxified_content = NULL;
    char* base_url = NULL;
    ap_filter_t* fnext = f->next ;
    apr_bucket_brigade* bbline ;
    
    if (f->r->main != NULL)  /*no subrequests*/
    {
        ap_remove_output_filter(f);
        return ap_pass_brigade(fnext, bb);
    }
    
    proxy_css_dir_conf* cfg
        = ap_get_module_config(f->r->per_dir_config, &proxy_css_module) ;

    proxy_css_ctx* ctx = f->ctx ;
    if (ctx == NULL)
    {
        /*unset content length as it may change and disable caching*/
        apr_table_unset(f->r->headers_out, "Content-Length") ;
        apr_table_unset(f->r->headers_out, "Content-MD5") ;
        apr_table_unset(f->r->headers_out, "ETag") ;

        ctx = f->ctx = apr_palloc(f->r->pool, sizeof(proxy_css_ctx)) ;
        ctx->bbsave = apr_brigade_create(f->r->pool, f->c->bucket_alloc) ;

        /*destroy input brigade and pass eos down the chain if no content type is present*/
        if(!f->r->content_type)
        {
            apr_brigade_destroy(bb);
            APR_BRIGADE_INSERT_TAIL(ctx->bbsave,
                apr_bucket_eos_create(ctx->bbsave->bucket_alloc) ) ;
            return ap_pass_brigade(f->next, ctx->bbsave) ;
        }

        ctx->htmlprefix = apr_table_get(f->r->subprocess_env, "psiphon_html_prefix");
        ctx->cssprefix = apr_table_get(f->r->subprocess_env, "psiphon_css_prefix");

        if (!ctx->htmlprefix || !*ctx->htmlprefix)
        {
            ap_remove_output_filter(f);
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r, "html prefix is empty.") ;
            return APR_EGENERAL;
        }
        if (!ctx->cssprefix || !*ctx->cssprefix)
        {
            ap_remove_output_filter(f);
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r, "css prefix is empty.") ;
            return APR_EGENERAL;
        }
    }

    /* bbline is what goes to the next filter,
     * so we (can) have a new one each time.
     */
    bbline = apr_brigade_create(f->r->pool, f->c->bucket_alloc) ;

    /* first ensure we have no mid-line breaks that might be in the
     * middle of a search string causing us to miss it!  At the same
     * time we split into lines to avoid pattern-matching over big
     * chunks of memory.
     */
    while ( b != APR_BRIGADE_SENTINEL(bb) )
    {
        if ( !APR_BUCKET_IS_METADATA(b) )
        {
            if ( apr_bucket_read(b, &buf, &bytes, APR_BLOCK_READ) == APR_SUCCESS )
            {
                if ( bytes == 0 )
                {
                    APR_BUCKET_REMOVE(b) ;
                }
                else while ( bytes > 0 )
                {
                    switch (cfg->lineend)
                    {

                        case LINEEND_UNIX:
                            le = memchr(buf, '\n', bytes) ;
                            break ;

                        case LINEEND_MAC:
                            le = memchr(buf, '\r', bytes) ;
                            break ;

                        case LINEEND_DOS:
                            /* Edge-case issue: if a \r\n spans buckets it'll get missed.
                             * Not a problem for present purposes, but would be an issue
                             * if we claimed to support pattern matching on the lineends.
                             */
                            found = 0 ;
                            le = memchr(buf+1, '\n', bytes-1) ;
                            while ( le && !found )
                            {
                                if ( le[-1] == '\r' )
                                {
                                    found = 1 ;
                                }
                                else
                                {
                                    le = memchr(le+1, '\n', bytes-1 - (le+1 - buf)) ;
                                }
                            }
                            if ( !found )
                                le = 0 ;
                            break;

                        case LINEEND_ANY:
                        case LINEEND_UNSET:
                            /* Edge-case notabug: if a \r\n spans buckets it'll get seen as
                             * two line-ends.  It'll insert the \n as a one-byte bucket.
                             */
                            le_n = memchr(buf, '\n', bytes) ;
                            le_r = memchr(buf, '\r', bytes) ;
                            if ( le_n != NULL )
                                if ( le_n == le_r + sizeof(char))
                                    le = le_n ;
                            else if ( (le_r < le_n) && (le_r != NULL) )
                                le = le_r ;
                            else
                                le = le_n ;
                            else
                                le = le_r ;
                            break;

                        case LINEEND_NONE:
                            le = 0 ;
                            break;

                        case LINEEND_CUSTOM:
                            le = memchr(buf, cfg->lechar, bytes) ;
                            break;
                    }
                    if ( le )
                    {
                        /* found a lineend in this bucket. */
                        offs = 1 + ((unsigned int)le-(unsigned int)buf) / sizeof(char) ;
                        apr_bucket_split(b, offs) ;
                        bytes -= offs ;
                        buf += offs ;
                        b1 = APR_BUCKET_NEXT(b) ;
                        APR_BUCKET_REMOVE(b);

                        /* Is there any previous unterminated content ? */
                        if ( !APR_BRIGADE_EMPTY(ctx->bbsave) )
                        {
                            /* append this to any content waiting for a lineend */
                            APR_BRIGADE_INSERT_TAIL(ctx->bbsave, b) ;
                            rv = apr_brigade_pflatten(ctx->bbsave, &fbuf, &fbytes, f->r->pool) ;
                            /* make b a new bucket of the flattened stuff */
                            b = apr_bucket_pool_create(fbuf, fbytes, f->r->pool,
                                f->r->connection->bucket_alloc) ;

                            /* bbsave has been consumed, so clear it */
                            apr_brigade_cleanup(ctx->bbsave) ;
                        }
                        /* b now contains exactly one line */
                        APR_BRIGADE_INSERT_TAIL(bbline, b);
                        b = b1 ;
                    }
                    else
                    {
                        /* no lineend found.  Remember the dangling content */
                        APR_BUCKET_REMOVE(b);
                        APR_BRIGADE_INSERT_TAIL(ctx->bbsave, b);
                        bytes = 0 ;
                    }
                }                /* while bytes > 0 */
            }
            else
            {
                /* bucket read failed - oops !  Let's remove it. */
                APR_BUCKET_REMOVE(b);
            }
        }
        else if ( APR_BUCKET_IS_EOS(b) )
        {
            /* If there's data to pass, send it in one bucket */
            if ( !APR_BRIGADE_EMPTY(ctx->bbsave) )
            {
                rv = apr_brigade_pflatten(ctx->bbsave, &fbuf, &fbytes, f->r->pool) ;
                b1 = apr_bucket_pool_create(fbuf, fbytes, f->r->pool,
                    f->r->connection->bucket_alloc) ;
                APR_BRIGADE_INSERT_TAIL(bbline, b1);
            }
            apr_brigade_cleanup(ctx->bbsave) ;
            /* start again rather than segfault if a seriously buggy
             * filter in front of us sent a bogus EOS
             */
            f->ctx = NULL ;

            /* move the EOS to the new brigade */
            APR_BUCKET_REMOVE(b);
            APR_BRIGADE_INSERT_TAIL(bbline, b);
        }
        else
        {
            APR_BUCKET_REMOVE(b);
            APR_BRIGADE_INSERT_TAIL(bbline, b);
        }
        /* OK, reset pointer to what's left (since we're not in a for-loop) */
        b = APR_BRIGADE_FIRST(bb) ;
    }

    /* OK, now we have a bunch of complete lines in bbline,
     * so we can apply our edit rules
     */

    for ( b = APR_BRIGADE_FIRST(bbline) ;
        b != APR_BRIGADE_SENTINEL(bbline) ;
        b = APR_BUCKET_NEXT(b) )
    {
        if ( !APR_BUCKET_IS_METADATA(b)
            && (apr_bucket_read(b, &buf, &bytes, APR_BLOCK_READ) == APR_SUCCESS))
        {
            if (f->r->proxyreq && f->r->filename
                && strncmp(f->r->filename, "proxy:", 6) == 0)
            {
                base_url = apr_pstrdup(f->r->pool, &f->r->filename[6]);
            }

            proxified_content = psi_proxify_css(
                                    f->r->pool,
                                    buf,
                                    bytes,
                                    base_url,
                                    ctx->htmlprefix,
                                    ctx->cssprefix);

            proxy_css_log("proxy_css in: %s", apr_pstrndup(f->r->pool, buf, bytes));
            proxy_css_log("proxy_css out: %s", proxified_content);

            b1 = APR_BUCKET_NEXT(b) ;
            apr_bucket_delete(b) ;

            b = apr_bucket_immortal_create(proxified_content,
                strlen(proxified_content), f->r->connection->bucket_alloc) ;

            APR_BUCKET_INSERT_BEFORE(b1, b) ;
        }
    }

    /* now pass it down the chain */
    rv = ap_pass_brigade(f->next, bbline) ;

    /* if we have leftover data, don't risk it going out of scope */
    for ( b = APR_BRIGADE_FIRST(ctx->bbsave) ;
        b != APR_BRIGADE_SENTINEL(ctx->bbsave) ;
        b = APR_BUCKET_NEXT(b))
    {
        apr_bucket_setaside(b, f->r->pool) ;
    }

    return rv ;
}


static void proxy_css_hooks(apr_pool_t* pool)
{
    ap_register_output_filter_protocol("proxy-css", proxy_css_filter,
        NULL, AP_FTYPE_RESOURCE,
        AP_FILTER_PROTO_CHANGE|AP_FILTER_PROTO_CHANGE_LENGTH) ;
    APR_REGISTER_OPTIONAL_FN(psiphon_get_css_prefix);
}


static const char* proxy_css_lineend(cmd_parms* cmd,
void* cfg, const char* arg, const char *ch)
{
    proxy_css_dir_conf* fcfg = cfg ;
    if (!strcasecmp(arg, "unix"))
    {
        fcfg->lineend = LINEEND_UNIX ;
    }
    else if (!strcasecmp(arg, "dos"))
    {
        fcfg->lineend = LINEEND_DOS ;
    }
    else if (!strcasecmp(arg, "mac"))
    {
        fcfg->lineend = LINEEND_MAC ;
    }
    else if (!strcasecmp(arg, "any"))
    {
        fcfg->lineend = LINEEND_ANY ;
    }
    else if (!strcasecmp(arg, "none"))
    {
        fcfg->lineend = LINEEND_NONE ;
    }
    else if (!strcasecmp(arg, "custom"))
    {
        if (ch)
        {
            fcfg->lineend = LINEEND_CUSTOM ;
            fcfg->lechar = ch[0];
        }
        else
        {
            return "You must specify the custom lineend character.";
        }
    }
    else
    {
        return "Unknown lineend scheme";
    }
    return NULL;
}

static void* proxy_css_cr_cfg(apr_pool_t* pool, char* x)
{
    proxy_css_dir_conf* ret = apr_palloc(pool, sizeof(proxy_css_dir_conf)) ;
    ret->lineend = LINEEND_UNSET;
    ret->lechar = 0;
    return ret ;
}


static void* proxy_css_dir_config(apr_pool_t* pool, char* dir)
{
    proxy_css_dir_conf* ret = apr_palloc(pool, sizeof(proxy_css_dir_conf)) ;
    ret->lineend = LINEEND_UNSET;
    ret->lechar = 0;
    ret->loc = dir;
    return ret ;
}

static void* proxy_css_srv_config(apr_pool_t * p, server_rec * x)
{
    proxy_css_srv_conf *cfg = apr_pcalloc(p, sizeof(proxy_css_srv_conf));
    cfg->module_location = NULL;
    return cfg;
}

static int register_module_location(apr_pool_t* pool, server_rec* s,
const char* location)
{
    proxy_css_srv_conf *conf =
        ap_get_module_config(s->module_config, &proxy_css_module);
    if(conf->module_location == NULL)
    {
        conf->module_location = apr_pstrdup(pool, location);
        return 0;
    }
    return 1;
}


static const char* set_proxy_css(cmd_parms *parms, void *in_dconf, int arg)
{
    proxy_css_dir_conf *dconf = in_dconf;
    dconf->proxy_css = arg;
    if(!dconf->loc)
    {
        return "proxy-css must be set in per-directory context only";
    }
    if(register_module_location(parms->pool, parms->server, dconf->loc) !=0 )
    {
        return "proxy-css must be set only once per server, no multiple locations allowed";
    }
    return NULL;
}

static const command_rec proxy_css_cmds[] =
{
    AP_INIT_FLAG("ProxyCSS", set_proxy_css, NULL, ACCESS_CONF, 
            "on/off flag"),
    AP_INIT_TAKE12("ProxyCSSLineEnd", proxy_css_lineend, NULL, OR_ALL,
            "Use line ending: UNIX|MAC|DOS|ANY|NONE|CUSTOM [char]") ,
    {NULL}
} ;

module AP_MODULE_DECLARE_DATA proxy_css_module =
{
    STANDARD20_MODULE_STUFF,
    proxy_css_dir_config,
    NULL,
    proxy_css_srv_config,
    NULL,
    proxy_css_cmds,
    proxy_css_hooks
};
