/*
    Psiphon Circumvention Platform
    Copyright (C) 2009 Psiphon Inc.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/********************************************************************
  Copyright (c) 2005-6, WebThing Ltd
  Author: Nick Kew <nick@webthing.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*********************************************************************/

/* apache */
#include <httpd.h>
#include <http_protocol.h>
#include <http_config.h>
#include <http_log.h>
#include <apr_strings.h>
#include <apr_xlate.h>
#include <apr_hash.h>

#include <apr_optional.h>

module AP_MODULE_DECLARE_DATA xml2enc_module;

#define DEFAULT_ENCODING "ISO-8859-1"
#define BUFLEN 65536
#define BUF_MIN 32768
#define APR_BRIGADE_DO(b,bb) for (b = APR_BRIGADE_FIRST(bb); \
        b != APR_BRIGADE_SENTINEL(bb); b = APR_BUCKET_NEXT(b))

#define ENC_INITIALISED 0x100
#define ENC_SEEN_EOS 0x200
#define ENC_SKIPTO ENCIO_SKIPTO

#define DEBUG
#undef DEBUG


#ifdef DEBUG
#define xm2enc_log(args...) syslog(LOG_DEBUG, args)
#else
#define xm2enc_log(args...)  ""
#endif

typedef enum {
    MODE_TO_DEFAULT, MODE_TO_ORIGINAL
}
xml2mode_t;

typedef struct
{
    const char* val;
} tattr;

typedef struct {
    char* buf;
    apr_size_t bytes;
    apr_xlate_t* convset;
    unsigned int flags;
    xml2mode_t mode;
    apr_off_t bblen;
    apr_bucket_brigade* bbnext;
    apr_bucket_brigade* bbsave;
    const char* encoding_from;
    const char* encoding_to;
} xml2ctx;

typedef struct {
    const char* env_charset_original;
    const char* default_charset;
    const char* original_charset;
//  const char* html_filter_name;
    apr_array_header_t* parsers;
} xml2cfg;

static ap_regex_t* seek_meta_ctype;
static ap_regex_t* seek_charset;
static apr_hash_t *iconv_enc_aliases;

static const char* set_parsers(cmd_parms* cmd, void* cnf, const char* arg)
{
    tattr* attr;
    xml2cfg* cfg = cnf;
    if (cfg->parsers == NULL)
        cfg->parsers = apr_array_make(cmd->pool, 3, sizeof(tattr));
    attr = apr_array_push(cfg->parsers) ;
    attr->val = arg;
    return NULL ;
}
 
static const char* lookup_iconv_enc(const char* enc) {
    const char* iconv_enc;
    iconv_enc = (const char*) apr_hash_get(iconv_enc_aliases, enc, APR_HASH_KEY_STRING);
    if (!iconv_enc)
        return (const char*) enc;
    return iconv_enc;
}

static const char* interpolate_vars(request_rec* r, const char* str) {
    const char* start;
    const char* end;
    const char* delim;
    const char* before;
    const char* after;
    const char* replacement;
    const char* var;
    for (;;) {

        start = str;

        if (!start || !*start)
            break;

        if (start = ap_strstr_c(start, "${"), start == NULL)
            break;

        if (end = ap_strchr_c(start + 2, '}'), end == NULL)
            break;

        delim = ap_strchr_c(start, '|');
        before = apr_pstrndup(r->pool, str, start - str);
        after = end + 1;
        if (delim) {
            var = apr_pstrndup(r->pool, start + 2, delim - start - 2);
        } else {
            var = apr_pstrndup(r->pool, start + 2, end - start - 2);
        }
        replacement = apr_table_get(r->subprocess_env, var);
        if (!replacement) {
            if (delim)
                replacement = apr_pstrndup(r->pool, delim + 1, end - delim - 1);
            else
                replacement = "";
        }
        str = apr_pstrcat(r->pool, before, replacement, after, NULL);
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                "Interpolating %s  =>  %s", var, replacement);
    }
    return str;
}

static int xml2enc_post_config(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s) {
    iconv_enc_aliases = apr_hash_make(p);
    apr_hash_set(iconv_enc_aliases, "gb2312", APR_HASH_KEY_STRING, "GBK");
    apr_hash_set(iconv_enc_aliases, "win-1251", APR_HASH_KEY_STRING, "WINDOWS-1251");
    return OK;
}

static char* sniff_BOM(unsigned char* in)
{
    /*
    *TODO: Add more charsets
    */
    if ((in[0] == 0xFE) && (in[1] == 0xFF))
        return("UTF-16");
    if ((in[0] == 0xFF) && (in[1] == 0xFE))
        return("UTF-16");
    return NULL;
}
static void sniff_encoding(request_rec* r, xml2ctx* ctx) {
    xml2cfg* cfg;
    char* p;
    apr_bucket* b;
    ap_regmatch_t match[2];
    apr_status_t rv;
    char* tmpenc = NULL;
    const char* ctype = r->content_type;
    const char* charset_str = NULL;

    cfg = ap_get_module_config(r->per_dir_config, &xml2enc_module);


    if (ctype) {
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "Content-Type is %s", ctype);
        if (ctype && (p = ap_strcasestr(ctype, "charset="), p != NULL)) {
            p += 8;
            if (tmpenc = apr_pstrndup(r->pool, p, strcspn(p, " ;")), tmpenc) {
                ap_str_tolower(tmpenc);
                ctx->encoding_from = lookup_iconv_enc((const char*) tmpenc);
                xm2enc_log("xml2enc: Got charset %s from HTTP headers", ctx->encoding_from);
            }
        }
    }

    tmpenc = sniff_BOM((unsigned char*)ctx->buf);
    if (tmpenc) {
        ctx->encoding_from = lookup_iconv_enc((const char*) tmpenc);
        if (ctx->encoding_from && *ctx->encoding_from) {
            xm2enc_log("xml2enc: Got charset %s from BOM", ctx->encoding_from);
        }
    }

    if (ap_regexec(seek_meta_ctype, ctx->buf, 1, match, 0) == 0) {

        p = apr_pstrndup(r->pool, ctx->buf + match[0].rm_so,
                match[0].rm_eo - match[0].rm_so);

        if (ap_regexec(seek_charset, p, 2, match, 0) == 0) {
            if (tmpenc = apr_pstrndup(r->pool, p + match[1].rm_so,
                    match[1].rm_eo - match[1].rm_so), tmpenc) {
                ap_str_tolower(tmpenc);
                ctx->encoding_from = lookup_iconv_enc((const char*) tmpenc);

                if (ctx->encoding_from && *ctx->encoding_from) {
                    xm2enc_log("xml2enc: Got charset %s from HTML META", ctx->encoding_from);
                }
            }
        }
    }
    switch (ctx->mode) {
        case MODE_TO_DEFAULT:
            xm2enc_log("xml2enc: ctx->mode = MODE_TO_DEFAULT");
            if (!ctx->encoding_from || !*ctx->encoding_from)
            {
                ctx->encoding_from = "ISO-8859-1";
            }
            ctx->encoding_to = cfg->default_charset;
            break;

        case MODE_TO_ORIGINAL:
            xm2enc_log("xml2enc: ctx->mode MODE_TO_ORIGINAL");
            ctx->encoding_from = cfg->default_charset;
            charset_str = interpolate_vars(r, cfg->original_charset);
            if (charset_str)
            {
                ctx->encoding_to = lookup_iconv_enc(charset_str);
            }
            else
            {
                ctx->encoding_to = DEFAULT_ENCODING;
            }

            break;
    }
    xm2enc_log("xml2enc: ctx->encoding_to = %s; ctx->encoding_from = %s", ctx->encoding_to, ctx->encoding_from);

    if (!apr_xlate_open(&ctx->convset, ctx->encoding_to, ctx->encoding_from, r->pool) == APR_SUCCESS) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                "Charset %s not supported.", ctx->encoding_from);
    }

    if (ctype && ctx->encoding_from) {
        if (cfg->env_charset_original && ctx->mode == MODE_TO_DEFAULT)
            apr_table_set(r->subprocess_env, cfg->env_charset_original, ctx->encoding_from);
    }
}

static apr_status_t xml2enc_filter_init(ap_filter_t* f) {
    xml2ctx* ctx;
    int i;
    ap_filter_t *ff = f->r->output_filters;
    xml2cfg* cfg = ap_get_module_config(f->r->per_dir_config, &xml2enc_module);

    if (!f->ctx) {
        f->ctx = ctx = apr_pcalloc(f->r->pool, sizeof (xml2ctx));

        ctx->mode = MODE_TO_DEFAULT;

        while (ff && ff->frec->name && cfg->parsers) {
            for (i=0; i < cfg->parsers->nelts; ++i) {
                tattr* attrs = (tattr*) cfg->parsers->elts;
            if (0 == strcmp(ff->frec->name, attrs[i].val)) {
                    /*
                    * there is a parser in front of us in the filters chain
                    */
                    ctx->mode = MODE_TO_ORIGINAL;
                    return APR_SUCCESS;
                }
            }
            if (ff == f) {
                /*
                * found itself in the filters chain
                */
                return APR_SUCCESS;
            }
            ff = ff->next;
        }
   }

   return APR_SUCCESS;
}

static apr_status_t xml2enc_ffunc(ap_filter_t* f, apr_bucket_brigade* bb) {
    xml2ctx* ctx = f->ctx;
    apr_status_t rv;
    apr_bucket* b;
    apr_bucket* bstart;
    apr_size_t insz = 0;

    if (!ctx) {
        /* log error about configuring this */
        ap_remove_output_filter(f);
        return ap_pass_brigade(f->next, bb);
    }

    if (ctx->bbsave == NULL) {
        ctx->bbsave = apr_brigade_create(f->r->pool, f->r->connection->bucket_alloc);
    }
    /* append to any data left over from last time */
    APR_BRIGADE_CONCAT(ctx->bbsave, bb);

    if (!(ctx->flags & ENC_INITIALISED)) {
        /* some kind of initialisation required */

        /* if we don't have enough data to sniff but more's to come, wait for it */
        rv = apr_brigade_length(ctx->bbsave, 0, &ctx->bblen);
        if ((ctx->bblen < BUF_MIN) && (ctx->bblen != -1)) {

            APR_BRIGADE_DO(b, ctx->bbsave) {
                if (APR_BUCKET_IS_EOS(b)) {
                    ctx->flags |= ENC_SEEN_EOS;
                    break;
                }
            }
            if (!(ctx->flags & ENC_SEEN_EOS)) {

                /* not enough data to sniff.  Wait for more */
                APR_BRIGADE_DO(b, ctx->bbsave) {
                    apr_bucket_setaside(b, f->r->pool);
                }
                return APR_SUCCESS;
            }
        }
        if (ctx->bblen == -1) {
            ctx->bblen = BUFLEN - 1;
        }
        /* flatten it into a NULL-terminated string */
        ctx->buf = apr_palloc(f->r->pool, ctx->bblen + 1);
        ctx->bytes = ctx->bblen;
        rv = apr_brigade_flatten(ctx->bbsave, ctx->buf, &ctx->bytes);
        ctx->buf[ctx->bytes] = 0;
        sniff_encoding(f->r, ctx);
        ctx->flags |= ENC_INITIALISED;
        ap_set_module_config(f->r->request_config, &xml2enc_module, ctx);
    }
    if (ctx->bbnext == NULL) {
        ctx->bbnext = apr_brigade_create(f->r->pool, f->r->connection->bucket_alloc);
    }

    if (!ctx->convset) {
        rv = ap_pass_brigade(f->next, ctx->bbsave);
        apr_brigade_cleanup(ctx->bbsave);
        ap_remove_output_filter(f);
        return rv;
    }
    /* move the data back to bb */
    APR_BRIGADE_CONCAT(bb, ctx->bbsave);

    while (b = APR_BRIGADE_FIRST(bb), b != APR_BRIGADE_SENTINEL(bb)) {
        ctx->bytes = 0;
        if (APR_BUCKET_IS_METADATA(b)) {
            if (APR_BUCKET_IS_EOS(b)) {
                /* send remaining data */
                //        return ap_fflush(f->next, ctx->bbnext);
                APR_BRIGADE_INSERT_TAIL(ctx->bbnext,
                        apr_bucket_eos_create(ctx->bbnext->bucket_alloc));
                ap_pass_brigade(f->next, ctx->bbnext);
            } else if (APR_BUCKET_IS_FLUSH(b)) {
                ap_fflush(f->next, ctx->bbnext);
            }
            APR_BUCKET_REMOVE(b);
            apr_bucket_destroy(b);
        } else { /* data bucket */
            char* buf;
            apr_size_t bytes = 0;
            char fixbuf[BUFLEN];
            apr_bucket* bdestroy = NULL;
            if (insz > 0) { /* we have dangling data.  Flatten it. */
                buf = fixbuf;
                bytes = BUFLEN;
                rv = apr_brigade_flatten(bb, buf, &bytes);
                if (bytes == insz) {
                    /* this is only what we've already tried to convert.
                     * The brigade is exhausted.
                     * Save remaining data for next time round
                     */

                    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
                            "xml2enc: Setting aside %" APR_SIZE_T_FMT
                            " unconverted bytes", bytes);
                    rv = ap_fflush(f->next, ctx->bbnext);
                    APR_BRIGADE_CONCAT(ctx->bbsave, bb);

                    APR_BRIGADE_DO(b, ctx->bbsave) {
                        apr_bucket_setaside(b, f->r->pool);
                    }
                    return rv;
                }
                /* remove the data we've just read */
                rv = apr_brigade_partition(bb, bytes, &bstart);
                while (b = APR_BRIGADE_FIRST(bb), b != bstart) {
                    APR_BUCKET_REMOVE(b);
                    apr_bucket_destroy(b);
                }
                ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r, "xml2enc: consuming %"
                        APR_SIZE_T_FMT " bytes flattened", bytes);
            } else {
                rv = apr_bucket_read(b, (const char**) & buf, &bytes, APR_BLOCK_READ);
                APR_BUCKET_REMOVE(b);
                bdestroy = b; /* can't destroy until we've finished with the data */
                ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r, "xml2enc: consuming %"
                        APR_SIZE_T_FMT " bytes from bucket", bytes);
            }
            /* OK, we've got some input we can use in [buf,bytes] */
            if (rv == APR_SUCCESS) {
                apr_size_t consumed;
                consumed = insz = bytes;
                while (insz > 0) {
                    if (ctx->bytes == ctx->bblen) {
                        /* nothing was converted last time!
                         * break out of this loop!
                         */
                        b = apr_bucket_transient_create(buf + (bytes - insz), insz,
                                bb->bucket_alloc);
                        APR_BRIGADE_INSERT_HEAD(bb, b);
                        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
                                "xml2enc: reinserting %" APR_SIZE_T_FMT
                                " unconsumed bytes from bucket", insz);
                        break;
                    }
                    ctx->bytes = ctx->bblen;
                    rv = apr_xlate_conv_buffer(ctx->convset, buf + (bytes - insz), &insz,
                            ctx->buf, &ctx->bytes);
                    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, rv, f->r,
                            "xml2enc: converted %" APR_SIZE_T_FMT "/%" APR_OFF_T_FMT " bytes",
                            consumed - insz, ctx->bblen - ctx->bytes);

                    consumed = insz;
                    ap_fwrite(f->next, ctx->bbnext, ctx->buf, ctx->bblen - ctx->bytes);
                    switch (rv) {
                        case APR_SUCCESS:
                            continue;
                        case APR_INCOMPLETE:
                            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r, "INCOMPLETE");
                            continue; /* If outbuf was too small, go round again.
                             * If it was inbuf, we'll break out when we test
                             * ctx->bytes == ctx->bblen
                             */
                        case APR_EINVAL: /* try skipping one bad byte */
                            ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, f->r,
                                    "Skipping invalid byte(s) in input stream!");
                            --insz;
                            --ctx->bytes; /*E.F seems to fix the bug with insert offset*/
                            continue;
                        default:
                            /* Erk!  What's this?
                             * Bail out, flush, and hope to eat the buf raw
                             */
                            ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, f->r,
                                    "Failed to convert input; trying it raw");
                            ctx->convset = NULL;
                            ap_fflush(f->next, ctx->bbnext);
                            return ap_pass_brigade(f->next, ctx->bbnext);
                    }
                }
            } else {
                ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, f->r,
                        "xml2enc: error reading data");
            }
            if (bdestroy) {
                apr_bucket_destroy(bdestroy);
            }
        }
    }
    return APR_SUCCESS;
}

#define PROTO_FLAGS AP_FILTER_PROTO_CHANGE|AP_FILTER_PROTO_CHANGE_LENGTH

static void xml2enc_hooks(apr_pool_t* pool) {
    ap_register_output_filter_protocol("xml2enc", xml2enc_ffunc,
            xml2enc_filter_init, AP_FTYPE_RESOURCE, PROTO_FLAGS);
    ap_hook_post_config(xml2enc_post_config, NULL, NULL, APR_HOOK_MIDDLE);

    seek_meta_ctype = ap_pregcomp(pool,
            "(<meta[^>]*http-equiv[ \t\r\n='\"]*content-type[^>]*>)",
            AP_REG_EXTENDED | AP_REG_ICASE);
    seek_charset = ap_pregcomp(pool, "charset=([A-Za-z0-9_-]+)",
            AP_REG_EXTENDED | AP_REG_ICASE);
}

static const command_rec xml2enc_cmds[] = {
    AP_INIT_TAKE1("xml2encDefaultCharset", ap_set_string_slot,
    (void*) APR_OFFSETOF(xml2cfg, default_charset),
    RSRC_CONF | ACCESS_CONF,
    "Usage: xml2encDefaultCharset charset"),
    AP_INIT_TAKE1("xml2encOriginalCharset", ap_set_string_slot,
    (void*) APR_OFFSETOF(xml2cfg, original_charset),
    RSRC_CONF | ACCESS_CONF,
    "Usage: xml2encOriginalCharset ${var} where var is env var name that holds original document encoding"),
    AP_INIT_TAKE1("xml2encEnvCharsetOriginal", ap_set_string_slot,
    (void*) APR_OFFSETOF(xml2cfg, env_charset_original),
    RSRC_CONF | ACCESS_CONF,
    "Usage: xml2encEnvCharsetOriginal name [of env var that holds original document encoding]"),
    AP_INIT_ITERATE("xml2encParsers", set_parsers, NULL,
    RSRC_CONF | ACCESS_CONF,
    "Usage: xml2encParsers list for post-parsing processing"),
   { NULL}
};

static void* xml2enc_config(apr_pool_t* pool, char* x) {
    xml2cfg* ret = apr_pcalloc(pool, sizeof (xml2cfg));
    return ret;
}

static void* xml2enc_merge(apr_pool_t* pool, void* BASE, void* ADD) {
    xml2cfg* base = BASE;
    xml2cfg* add = ADD;
    xml2cfg* ret = apr_pcalloc(pool, sizeof (xml2cfg));
    ret->env_charset_original = add->env_charset_original ? add->env_charset_original : base->env_charset_original;
    ret->default_charset = add->default_charset ? add->default_charset : base->default_charset;
    ret->original_charset = add->original_charset ? add->original_charset : base->original_charset;
    ret->parsers = add->parsers ? add->parsers : base->parsers;
    return ret;
}

module AP_MODULE_DECLARE_DATA xml2enc_module = {
    STANDARD20_MODULE_STUFF,
    xml2enc_config,
    xml2enc_merge,
    NULL,
    NULL,
    xml2enc_cmds,
    xml2enc_hooks
};
