#!/usr/sbin/php

<?php
/*
    Psiphon Circumvention Platform
    Copyright (C) 2009 Psiphon Inc.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 * Execute with no arguments to see usage.
 */

/*
 * Input file format requirements:
 *  - First column must be StringIDs. They do not need to be identical to the set
 *    of IDs in lang(_bb).php. (Super-, sub-, or not-even-intersecting set.)
 *  - First row must be: StringID, first-lang-id, second-lang-id, etc.
 *  - If a column of English strings (i.e., the ones provided to the translator)
 *    are present, they will be used to detect stale strings.
 */

// lang.php includes lang_bb.php and needs to know where to find it.
$_SERVER[DOCUMENT_ROOT] = "../../www";
include($_SERVER[DOCUMENT_ROOT]."/includes/lang.php");
@include($_SERVER[DOCUMENT_ROOT]."/admin/lang_admin.php");

if (count($argv) != 2)
{
    print("Usage: ./csv2lang.php {csvfilename}\n  Outputs results to stdout and updates lang files.\n");
    exit(0);
}

// Input parameters
$filename = $argv[1];
$allow_stale = true;
$allow_defunct = true;
$allow_missing = true;
$allow_english_copies = true;
$allow_bad_html_tags = true;
$allow_bad_format_specifiers = true;


// Reads past the UTF-8 bom if it is there.
// fgetcsv does not seem to like the presence of the BOM.
// Taken from: http://www.php.net/manual/en/function.fopen.php#78308
function fopen_utf8 ($filename, $mode)
{
    $file = @fopen($filename, $mode);
    $bom = fread($file, 3);
    if ($bom != b"\xEF\xBB\xBF")
        rewind($file);
    return $file;
}

// Open the CSV file
$fcsv = fopen_utf8($filename, "r");

// Read the languages present in the CSV
$entry = fgetcsv($fcsv);

if (!$entry)
{
    exit("File not found or otherwise bad (fgetcsv returned false): ".$filename);
}

if (count($entry) < 2 || $entry[0] != "StringID")
{
    exit("First line of file must be StringID,langcode1[,langcode2,...]");
}

$langs_in_csv = array_slice($entry, 1);
$langs_in_csv = array_filter($langs_in_csv); // remove blank entries created by empty columns

// Will be used to store the StringIDs where the $config English doesn't match the CSV English (if provided)
$stale_ids = array();
// Will be used to store the StringIDs where there is an entry in the CSV but not in $config.
$defunct_ids = array();
// Will be used to store the StringIDs where there is an entry in $config but not in the CSV.
$missing_ids = array_fill_keys(array_keys($config['msg']), "*");
// Will be used to store the StringIDs where the translation is identical to the English.
$english_copies_ids = array();
// Will be used to store the StringIDs where the order of the special tags deviates from the English.
$bad_html_tags_ids = array();
$bad_format_specifiers_ids = array();

while ($entry = fgetcsv($fcsv))
{
    $string_id = $entry[0];

    if (!$config['msg'][$string_id])
    {
        // If the current translation is defunct, just skip it.
        $defunct_ids[$string_id] = "*";
        continue;
    }

    // Remove the current StringID from $missing_ids (well, set it to null)
    $missing_ids[$string_id] = Null;

    $i = 1;
    foreach ($langs_in_csv as $lang_id)
    {
        $translation = $entry[$i];
        $i++;

        if ($lang_id == "en")
        {
            // Check for English mismatches
            if ($config['msg'][$string_id]['en'] != $translation)
            {
                $stale_ids[$string_id] .= $lang_id." ";
            }

            // Don't modify the English in the lang file.
            continue;
        }
        else // not English
        {
            // If the translation is empty, use the English.
            if (is_null($translation) || !strlen($translation))
            {
                $translation = $config['msg'][$string_id]['en'];
            }

            // Check for non-translated strings
            if (strtoupper($config['msg'][$string_id]['en']) == strtoupper($translation))
            {
                $english_copies_ids[$string_id] .= $lang_id." ";
            }

            // Check for HTML tags mismatches
            if (!compare_html_tags($translation, $config['msg'][$string_id]['en']))
            {
                $bad_html_tags_ids[$string_id] .= $lang_id." ";
            }

            // Check for format specifiers mismatches
            if (!compare_format_specifiers($translation, $config['msg'][$string_id]['en']))
            {
                $bad_format_specifiers_ids[$string_id] .= $lang_id." ";
            }
        }

        $config['msg'][$string_id][$lang_id] = $translation;
    }
}

// Warn about lang array
print(str_repeat("=", 80)."\n");
print("NOTE: If this is a totally new language, make sure to add it to \$config['lang']\n");

// Report defunct IDs
if (count($defunct_ids))
{
    print(str_repeat("=", 80)."\n");

    print("IDs present in CSV but not in \$config:\n");
    print(format_ids_array($defunct_ids));
    print("(NOTE: If these don't look like IDs, the CSV is probably broken.\n");
    print("Check if there are any instances of \\\" and replace with just \")\n");

    if (!$allow_defunct)
    {
        exit("Defunct StringIDs disallowed");
    }
}

// Report missing IDs
// $missing_ids has a null element for every non-missing ID, so we need to remove those.
$missing_ids = array_filter($missing_ids);
if (count($missing_ids))
{
    print(str_repeat("=", 80)."\n");

    print("IDs present in \$config but not in CSV:\n");
    print(format_ids_array($missing_ids));

    if (!$allow_missing)
    {
        exit("Missing StringIDs disallowed");
    }

    // To keep lang.php consistent, for every missing translation we'll put a stub
    // that is simply the English.

    foreach ($missing_ids as $id => $x)
    {
        foreach ($langs_in_csv as $lang_id)
        {
            if (!isset($config['msg'][$id][$lang_id]))
            {
                $config['msg'][$id][$lang_id] = $config['msg'][$id]['en'];
            }
        }
    }
}

// Report IDs of strings that are just copies of the English
if (count($english_copies_ids))
{
    print(str_repeat("=", 80)."\n");

    print("IDs where CSV translation is simply the English text:\n");
    print("(NOTE: May also indicate empty translations that are stubbed with English).\n");
    print(format_ids_array($english_copies_ids));

    if (!$allow_english_copies)
    {
        exit("Stale StringIDs disallowed");
    }
}

// Report stale IDs
if (count($stale_ids))
{
    print(str_repeat("=", 80)."\n");

    print("IDs where CSV English does not match \$config English (probably stale):\n");
    print(format_ids_array($stale_ids));

    if (!$allow_stale)
    {
        exit("Stale StringIDs disallowed");
    }
}

// Report HTML tag mismatch IDs
if (count($bad_html_tags_ids))
{
    print(str_repeat("=", 80)."\n");

    print("IDs where CSV HTML tags don't match current English (may not be fatal):\n");
    print(format_ids_array($bad_html_tags_ids));

    if (!$allow_bad_html_tags)
    {
        exit("HTML tag mismatches disallowed");
    }
}

// Report format specifiers mismatch IDs
if (count($bad_format_specifiers_ids))
{
    print(str_repeat("=", 80)."\n");

    print("IDs where CSV format specifiers don't match current English (PROBABLY FATAL):\n");
    print(format_ids_array($bad_format_specifiers_ids));

    if (!$allow_bad_format_specifiers)
    {
        exit("Format specifier mismatches disallowed");
    }
}

print(str_repeat("=", 80)."\n\n");

//
// At this point we have a fully updated $config and we've output our warnings.
// Output translations in the format used in lang.php. Example:
/*
    'language' => array(
        'en' => "Language",
        'fa' => "زبان",
        'ru' => "Язык",
    ),
 */
// lang_bb.php and lang_admin.php format:
/*
$config['msg']['bb_profile'] = array(
    'en' => "Profile",
    'ru' => "&#1053;&#1072;&#1089;&#1090;&#1088;&#1086;&#1081;&#1082;&#1080;",
    'fa' => "&#1582;&#1604;&#1575;",
);
 */

$lang_output = "";
$lang_bb_output = "";
$lang_psiadmin_output = "";
$expected_translation_count = Null;
$indent = "    ";
foreach($config['msg'] as $string_id => $translations)
{
    if (is_bb_string_id($string_id)) // lang_bb.php
    {
        $lang_bb_output .= "\$config['msg']['".$string_id."'] = array(\n";

        foreach($translations as $lang_id => $translation)
        {
            $translation = escape_translation($translation);

            $lang_bb_output .= $indent."'".$lang_id."' => \"".$translation."\",\n";
        }

        $lang_bb_output .= ");\n\n";
    }
    else if (is_psiadmin_string_id($string_id)) // lang_admin.php
    {
        $lang_psiadmin_output .= "\$config['msg']['".$string_id."'] = array(\n";

        foreach($translations as $lang_id => $translation)
        {
            $translation = escape_translation($translation);

            $lang_psiadmin_output .= $indent."'".$lang_id."' => \"".$translation."\",\n";
        }

        $lang_psiadmin_output .= ");\n\n";
    }
    else // lang.php
    {
        $lang_output .= $indent."'".$string_id."' => array(\n";

        foreach($translations as $lang_id => $translation)
        {
            // Escape all the double-quotes (and other escaped things) in the string.
            $translation = escape_translation($translation);

            $lang_output .= $indent.$indent."'".$lang_id."' => \"".$translation."\",\n";
        }

        $lang_output .= $indent."),\n\n";
    }

    // Sanity check to make sure we have a translation or stub for every StringID.
    if (is_null($expected_translation_count))
    {
        $expected_translation_count = count($translations);
    }
    else if (count($translations) != $expected_translation_count)
    {
        echo "$string_id => ";print_r($translations);echo "\n";
        exit("TRANSLATION COUNT MISMATCH\n");
    }
}

//
// Read and modify the contents of lang.php and lang_bb.php
//

$lang_files = array(
                $_SERVER[DOCUMENT_ROOT]."/includes/lang.php" => $lang_output,
                $_SERVER[DOCUMENT_ROOT]."/includes/lang_bb.php" => $lang_bb_output);
if ($lang_psiadmin_present)
{
    $lang_files[$_SERVER[DOCUMENT_ROOT]."/admin/lang_admin.php"] = $lang_psiadmin_output;
}
else if (strlen($lang_psiadmin_output) > 0)
{
    exit("ADMIN STRINGS TRANSLATED BUT lang_admin.php NOT INCLUDED\n");
}

foreach ($lang_files as $filename => $translation_block)
{
    $file = fopen($filename, "r");
    if (!$file) exit("Can not open $filename for reading");

    $file_contents = "";
    $in_translation_block = false;
    while ($line = fgets($file))
    {
        if (stristr($line, "// DO NOT MODIFY THIS LINE -- START OF TRANSLATIONS"))
        {
            $in_translation_block = true;
            $file_contents .= $line."\n";
            $file_contents .= $translation_block;
        }
        else if (stristr($line, "// DO NOT MODIFY THIS LINE -- END OF TRANSLATIONS"))
        {
            $in_translation_block = false;
            $file_contents .= $line;
        }
        else if ($in_translation_block)
        {
            // Do nothing -- throw out line
        }
        else
        {
            $file_contents .= $line;
        }
    }

    fclose($file);

    $file = fopen($filename, "w");
    if (!$file) exit("Can not open $filename for writing");

    if (fwrite($file, $file_contents) != strlen($file_contents))
    {
        exit("Can not write complete contents of $filename");
    }

    fclose($file);
}

// Escape all the double-quotes (and other escaped things) in the string.
function escape_translation($str)
{
    return addcslashes($str, "\\\"");
}

function is_bb_string_id($string_id)
{
    return (strncmp($string_id, "bb_", strlen("bb_")) == 0);
}

function is_psiadmin_string_id($string_id)
{
    return (strncmp($string_id, "psiadmin_", strlen("psiadmin_")) == 0);
}

function find_all_html_tag_positions($str)
{
    // Strip newlines (because I don't know how to ignore them in regular expressions)
    $str = str_replace(array("\r", "\n"), " ", $str);

    $special_positions = array();

    // Find all of the html open tags
    $pos = 0;
    while ($pos < strlen($str))
    {
        preg_match("/(<[a-zA-Z]+[\w]*)/", substr($str, $pos), $matches);
        $match_pos = strpos($str, $matches[0], $pos);
        if ($match_pos === FALSE) break;
        $special_positions[$match_pos] = $matches[0];
        $pos = $match_pos+2;
    }

    // Find all of the html close tags
    $pos = 0;
    while ($pos < strlen($str))
    {
        preg_match("/(<\/[a-zA-Z]+[\w]*)/", substr($str, $pos), $matches);
        $match_pos = strpos($str, $matches[0], $pos);
        if ($match_pos === FALSE) break;
        $special_positions[$match_pos] = $matches[0];
        $pos = $match_pos+2;
    }

    ksort($special_positions);

    return $special_positions;
}

function find_all_format_specifier_positions($str)
{
    // Strip newlines (because I don't know how to ignore them in regular expressions)
    $str = str_replace(array("\r", "\n"), " ", $str);

    $special_positions = array();

    // Find all of the format specifiers
    $pos = 0;
    while ($pos < strlen($str))
    {
        preg_match("/(%[a-z])/", substr($str, $pos), $matches);
        $match_pos = strpos($str, $matches[0], $pos);
        if ($match_pos === FALSE) break;
        $special_positions[$match_pos] = $matches[0];
        $pos = $match_pos+2;
    }

    ksort($special_positions);

    return $special_positions;
}

function compare_special_tag_arrays($array1, $array2)
{
    // They keys aren't important to the comparison, so just compare the values.
    return array_values($array1) === array_values($array2);
}

function compare_html_tags($str1, $str2)
{
    return compare_special_tag_arrays(
            find_all_html_tag_positions($str1),
            find_all_html_tag_positions($str2));
}

function compare_format_specifiers($str1, $str2)
{
    return compare_special_tag_arrays(
            find_all_format_specifier_positions($str1),
            find_all_format_specifier_positions($str2));
}

function format_ids_array($ids_array)
{
    $return = "";
    foreach ($ids_array as $string_id => $lang_ids)
    {
        $return .= $string_id." (".rtrim($lang_ids).")\n";
    }

    return $return;
}

exit(0);

?>
