<?
/*
    Psiphon Circumvention Platform
    Copyright (C) 2009 Psiphon Inc.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// create_account.php - external users create an account
//
// This module provides the functionality for 3 similar screens:
// 1. w.php: users click an invite link to create an account
// 2. create.php: users create an account and need to input a CAPCHA
// 3. anon_create.php: users create an account from an anonymous
//    browsing session (inherits home page attributes)
//
// These are all combined into one module since there is a lot of
// common code for each case.  The individual PHP files named above
// just set a "mode" flag and include create_account.php
//
// Modes: create_account_mode_anonymous, ..._invitation, ..._create

include_once($_SERVER[DOCUMENT_ROOT]."/includes/common_includes.php");
include_once($_SERVER[DOCUMENT_ROOT]."/includes/cr.php");
include_once($_SERVER[DOCUMENT_ROOT]."/includes/geoip_helpers.php");

// mode sanity check

if (!in_array($create_account_mode,
            array(create_account_mode_anonymous,
                  create_account_mode_invitation,
                  create_account_mode_create)) )
{
    syslog(LOG_DEBUG, "invalid mode in create_account");
    exit(0);
}

// redirect to SSL connection

if ($_SERVER["HTTPS"] != "on")
{
   header("HTTP/1.1 301 Moved Permanently");
   header("Location: https://".$_SERVER["SERVER_NAME"].$_SERVER["REQUEST_URI"]);
   exit(0);
}

//
// Business logic
//

// create PHP session for CAPCHA (see image-code.php)

session_start();

// Set invitation code from GET parameter in invite URL

if (!strlen($_POST[create_account_invitation_code]))
{
    $_POST[create_account_invitation_code] = $_GET[p];
}

// If the invitation code came from GET, it should be read-only.
$invitation_code_readonly = ($_GET[p] || $_POST['invitation_code_readonly']);

// Set form language
//
// Note one tricky order dependency: the $record_user[last_lang] is set in the
// anon and invite cases based on what's in the database and this affects the
// ___() output for the following validations.  That's why we check the anon
// session and/or invite code here for language, but don't start validating
// the form until later on when it's submitted.
// So, some of the following is validation, or half validation even though
// it's not in the main validation loop.

if (strlen($_POST["create_account_lang"]) && array_key_exists($_POST["create_account_lang"], $config["lang"]))
{
    // Keep the language setting the user specified and use it when re-rendering
    // the screen (___(), etc., which reference $record_user[last_lang])

    $record_user["last_lang"] = $_POST["create_account_lang"];
}

// By default, invite and anonymous forms use invite and anonymous
// session languages, respectively

if ($create_account_mode == create_account_mode_anonymous)
{
    // validation: check that this is an anonymous session
    // TODO: in other cases, should we deny access if already existing session?
    //       e.g., should regular, logged-in user be able to do create?
    //
    // this is our only anonymous session check, so this case
    // is run regardless of whether $_POST[create_account_lang] is already
    // set to something

    $is_anonymous_session = false;

    if ($_COOKIE[p])
    {
        // NOTE: get_anonymous_user_class_id() is not user input and does not need to be parameterized.
        $anonymous_user_record = lookup_user(
                                    $config,
                                    "session = :session_id and grp = '".get_anonymous_user_class_id()."'",
                                    array(":session_id" => $_COOKIE[p]));

        if ($anonymous_user_record)
        {
            if (!$_POST[create_account_lang]) $record_user[last_lang] = $anonymous_user_record[last_lang];
            $is_anonymous_session = true;
        }
    }

    if (!$is_anonymous_session)
    {
        // user should not be here at all, so no nice error messages
        include($_SERVER[DOCUMENT_ROOT]."/http-errors/404.php");
    }
}
elseif ($create_account_mode == create_account_mode_invitation)
{
    // NOT validation: load the invitation record and set the language,
    // but if not found, don't fail, just fall through to validation
    // loop below (in POST case)
    //
    // the loaded invite record is also used down below, so this case
    // is run regardless of whether $_POST[create_account_lang] is already
    // set to something
    //
    // Also ensure that invitation has not yet expired (in case not yet
    // been cleaned up by cron job).

    $_POST[create_account_invitation_code] = trim($_POST[create_account_invitation_code]);
    $current_proxy = (lookup_current_proxy($config));
    $query =
        "SELECT * FROM invitation ".
        "WHERE pass = :pass AND ".
        "proxy = :proxy AND ".
            "(date + '{$config['invitation_ttl']}') > UNIX_TIMESTAMP()"; // don't need to parameterize the config value

    $result = db_query_execute($config, $query, 
            convert_null_array(
                array(
                    ":pass" => $_POST['create_account_invitation_code'],
                    ":proxy" => $current_proxy
                    )
                ));

    $invitation_record = db_fetch_result($config, $result);
    if ($invitation_record)
    {
        if (!$_POST['create_account_lang']) $record_user['last_lang'] = $invitation_record['lang'];
    }
    else
    {
        // user should not be here at all, so no nice error messages
        include($_SERVER[DOCUMENT_ROOT]."/http-errors/404.php");
    }
    
    // display email inviation was sent to as default email for new account
    $default_email = $invitation_record['email'];
}


$privacy_info = '';
if ($create_account_mode == create_account_mode_anonymous)
{
    $privacy_info = sprintf(___("IMPORTANT PRIVACY INFORMATION:<br>
                <ul>
                    <li><b>Username and E-mail address:&nbsp;</b>Your username and e-mail address are visible only to Psiphon administrators. By providing an e-mail address, Psiphon can inform you of new ways to connect if necessary. We recommend you create a GMail (<a href=\"https://mail.google.com\">https://mail.google.com/</a>) e-mail account for this purpose, since it is web-based and simple to use. It is highly recommended that you provide an e-mail address, but it is not required in order to create a Psiphon account.</li>
                    <li style='margin-top:5px'><b>Password:&nbsp;</b>Your password is not visible to any Psiphon administrator or user. Use this password to log in to your Psiphon account. It is recommended that you follow best practices for online account password creation and that you do not use the same password that you use for any other online accounts.</li>
                </ul>
                For more information, please visit our <a href=\"%s\">Resource Centre</a>."), make_proxied_url($config["psiphon_web_resource_page"]));
}
else
{
    $privacy_info = ___("IMPORTANT PRIVACY INFORMATION:<br>
                <ul>
                    <li><b>Username and E-mail address:&nbsp;</b>Your username and e-mail address are visible only to Psiphon administrators. By providing an e-mail address, Psiphon can inform you of new ways to connect if necessary. We recommend you create a GMail (<a href=\"https://mail.google.com\">https://mail.google.com/</a>) e-mail account for this purpose, since it is web-based and simple to use. It is highly recommended that you provide an e-mail address, but it is not required in order to create a Psiphon account.</li>
                    <li style='margin-top:5px'><b>Password:&nbsp;</b>Your password is not visible to any Psiphon administrator or user. Use this password to log in to your Psiphon account. It is recommended that you follow best practices for online account password creation and that you do not use the same password that you use for any other online accounts.</li>
                </ul>
                For more information, please visit our Resource Centre.");
}

// email verification: all modes require it except invite, where the user got his
// invite via email and supplies the same email when creating his account.

$require_email_verification = true;

$curs = "create_account.create_account_username";

// process create account form

while ($_POST[create_account_submit] == "1")
{
    // validation flag: check all fields and output errors for each at once

    $valid = true;

    // check invitation code (also fetch invitation_record for use below)

    if ($create_account_mode == create_account_mode_invitation)
    {
        // Make sure user supplied correct code.  $invitation_record
        // was already loaded above, if it's valid.

        if (!$invitation_record)
        {
            $processing_errors[] = ___("Wrong invitation code");
            if ($valid) $curs="create_account.create_account_invitation_code";
            $valid = false;
        }
    }

    // check CAPTCHA

    if ($create_account_mode == create_account_mode_create
        || $create_account_mode == create_account_mode_anonymous)
    {
        // "code" is the session variable set by image-code.php
        // also, see image-code.php reference below
        $_POST[create_account_security_code] = trim($_POST[create_account_security_code]);
        if (!strlen($_POST[create_account_security_code]) ||
            ($_POST[create_account_security_code] != $_SESSION[code]))
        {
            $processing_errors[] = ___("Error: Wrong security code");
            $curs = "create_account.create_account_security_code";
            $valid = false;
        }
       //Making sure CAPTCHA session is not re-used
        unset($_SESSION[code]);
    }

    // check lang

    if (!in_array($_POST[create_account_lang], array_keys($config[lang])))
    {
         $processing_errors[] = ___("Error: No language selected");
         $valid = false;
    }

    // check username

    $_POST[create_account_username] = trim($_POST[create_account_username]);
    if (!strlen($_POST[create_account_username]) || !is_user($_POST[create_account_username]))
    {
        $processing_errors[] = ___("Error: Invalid username format");
        if ($valid) $curs = "create_account.create_account_username";
        break;
    }

    // check email (blank or valid format)

    $_POST[create_account_email] = trim($_POST[create_account_email]);
    if (strlen($_POST[create_account_email]) &&
        !is_email($_POST[create_account_email])
    )
    {
         $processing_errors[] = ___("Error: Invalid e-mail format");
         if ($valid) $curs = "create_account.create_account_email";
         $valid = false;
    }

    // email verfication exception for invite mode

    if ($create_account_mode == create_account_mode_invitation
        && $invitation_record
        && $invitation_record[email] == $_POST[create_account_email])
    {
        $require_email_verification = false;
    }

    // check password length and password verification

    if (!is_valid_password($config, $_POST[create_account_password], $_POST[create_account_username]))
    {
        $processing_errors[] = sprintf(___("Error: Password should be at least %d characters long and not contain your username"), $config[min_pass_length]);
        if ($valid) $curs = "create_account.create_account_password";
        $valid = false;
    }

    if ($_POST[create_account_password] != $_POST[create_account_verify_password])
    {
        $processing_errors[] = ___("Error: Password mismatch");
        if ($valid) $curs = "create_account.create_account_password";
        $valid = false;
    }

    // done validating inputs

    if (!$valid)
    {
        break;
    }

    // Assemble fields for new user record

    // determine user group
    //
    // 1. anonymous and create users get group/class "User"
    // 2. invite users get group specified in invite

    if ($create_account_mode == create_account_mode_invitation)
    {
        $user_group = $invitation_record[grp];
    }
    else // create_account_mode_invitation, create_account_mode_create
    {
        $user_group = get_normal_user_class_id(); // User
    }

    // determine proxy
    //
    // 1. anonymous user gets assigned to same proxy as anonymous session
    // 2. invite user gets assigned to proxy specified in invite
    // 3. create user gets assigned to current proxy, or fails if not on a proxy
    //
    // Note: there used to be an explict case for $config["no_proxy"], do
    // we still need to support that mode?

    $proxy_id = 0;

    if ($create_account_mode == create_account_mode_anonymous)
    {
        $proxy_id = $anonymous_user_record[current_proxy];
    }
    elseif ($create_account_mode == create_account_mode_invitation)
    {
        $proxy_id = $invitation_record[proxy];
    }
    else // create_account_mode_create
    {
        $query = "SELECT id FROM proxy WHERE hostname = :hostname";

        $result = db_query_execute($config, $query, convert_null_array(array(":hostname" => getenv("REMOTE_ADDR"))));

        $record = db_fetch_result($config, $result);
        if ($record)
        {
            $proxy_id = $record['id'];
        }
        else
        {
            syslog(LOG_DEBUG, getenv("HTTP_HOST").getenv("REQUEST_URI").": client ".getenv("REMOTE_ADDR").": no proxy found");
            include($_SERVER[DOCUMENT_ROOT]."/http-errors/403.php");
        }
    }

    // determine home_page and home_page_locked
    //
    // 1. anonymous user inherits home_page/locked from anonymous session
    // 2. invite and create users have default (blank) settings.  No local
    //    variable is set and SQL DEFAULT is inserted below.

    if ($create_account_mode == create_account_mode_anonymous)
    {
        $home_page = $anonymous_user_record[home_page];
        $home_page_locked = $anonymous_user_record[home_page_locked];
    }

    // determine invited date
    //
    // 1. invite user has invited (date) set to invite date.
    // 2. other users have default (blank) settings.  No local
    //    variable is set and SQL DEFAULT is inserted below.

    if ($create_account_mode == create_account_mode_invitation)
    {
        $invited = $invitation_record[date];
    }

    // GeoIP lookup

    $login_region = lookup_region($config);

    // password

    $pass_salt = bin2hex(secure_rand());
    $password = sha1($_POST["create_account_password"].$pass_salt);

    // unguessable session ID

    $session_id = bin2hex(secure_rand());

    // unguessable email verification token

    $email_candidate_token = bin2hex(secure_rand());

    // insert new user record
    //
    // email logic:
    // - empty email is always allowed, and NULL is inserted; no verfification
    // - if invitation was sent via email and user enters same email address,
    //   that email is immediately activated
    // - otherwise, verification (email set to NULL, email_candidate set to
    //   user supplied email, verification email sent)

    $activate_email = (strlen($_POST['create_account_email']) && !$require_email_verification) ;
    $verify_email = (strlen($_POST['create_account_email']) && $require_email_verification) ;

    // NOTE: Because we don't have transactions, there is a race condition between
    // the conflict check and the insert/update attempt. If a conflict is
    // introduced between the check and the insert/update, a DB error will occur.
    // This is unfortunate but acceptable (at this time).

    if (db_check_if_value_exists(
            $config,
            "user",
            "uname",
            $_POST['create_account_username']))
    {
        $processing_errors[] = ___("Username already exists in the system");
        $curs = "create_account.create_account_username";
        break;
    }

    if ($activate_email
        && db_check_if_value_exists(
            $config,
            "user",
            "email",
            $_POST['create_account_email']))
    {
        $processing_errors[] = ___("E-mail address already exists in the system");
        $curs = "create_account.create_account_email";
        break;
    }

    // Set default homepage
    $default_homepage_url = $invitation_record['homepage'];
    if ( is_url($default_homepage_url))
        $home_page = $default_homepage_url;

    $query = "INSERT INTO user (".
                "email,".
                "pass,".
                "pass_salt,".
                "uname,".
                "session,".
                "last_lang,".
                "last_reg,".
                "grp,".
                "current_proxy,".
                "created,".
                "invited,".
                "home_page,".
                "last,".
                "last_activity,".
                "home_page_locked, ".
                "email_candidate, ".
                "email_candidate_token, ".
                "email_candidate_date) ".
                "VALUES (";
    $query .=   ($activate_email ? ":email" : "NULL").",".
                ":pass,".
                ":pass_salt,".
                ":uname,".
                ":session,".
                ":last_lang,".
                ":last_reg,".
                ":grp,".
                ":current_proxy,".
                "UNIX_TIMESTAMP(),". // created = now
                ($invited_time ? ":invited_time" : "DEFAULT").",".
                ($home_page ? ":home_page" : "DEFAULT").",".
                "UNIX_TIMESTAMP(),". // last login = now
                "UNIX_TIMESTAMP(),". // last activity = now
                ($home_page_locked ? ":home_page_locked" : "DEFAULT").",".
                ($verify_email ? ":email_candidate" : "NULL").",".
                ($verify_email ? ":email_candidate_token" : "NULL").",".
                ($verify_email ? "UNIX_TIMESTAMP()" : "NULL").
                ")";

    db_query_execute(
        $config,
        $query,
        convert_null_array(array(
            ":pass" => $password,
            ":pass_salt" => $pass_salt,
            ":uname" => $_POST['create_account_username'],
            ":session" => $session_id,
            ":last_lang" => $record_user['last_lang'],
            ":last_reg" => $login_region,
            ":grp" => $user_group,
            ":current_proxy" => $proxy_id)
            + ($activate_email ? array(":email" => $_POST['create_account_email']) : array())
            + ($invited_time ? array(":invited_time" => $invited_time) : array())
            + ($home_page ? array(":home_page" => $home_page) : array())
            + ($home_page_locked ? array(":home_page_locked" => $home_page_locked) : array())
            + ($verify_email ? array(":email_candidate" => $_POST['create_account_email']) : array())
            + ($verify_email ? array(":email_candidate_token" => $email_candidate_token) : array())));

    $user_id = db_get_last_insert_id($config);

    // Insert the user-proxy relation
    $query = "INSERT INTO userproxy (userid, proxyid) ".
                "VALUES ((SELECT LAST_INSERT_ID()), :proxyid)";
    $query_params = array(":proxyid" => $proxy_id);
    db_query_execute($config, $query, convert_null_array($query_params));

    // Set "p" cookie to new session ID.
    // We're *not* setting the "cu" cookie, the URL encoding key.  Once
    // auto-URL encoding is implemented, it will be set later.

    header("Set-Cookie: p={$session_id}; path=/", false);

    // In the invitation case, the invitation is deleted as it has been consumed.
    // We also update stats for inviter.
    // Note: since there's no atomic transaction, there's a small chance that an
    // invitation could be used twice and that stats won't be updated.  This order
    // of things seems preferable to a small chance that an invitation is deleted
    // before the user is created.

    if ($create_account_mode == create_account_mode_invitation)
    {
        $query = "DELETE FROM invitation WHERE id = :invitation_id";
        db_query_execute($config, $query, convert_null_array(array(":invitation_id" => $invitation_record['id'])));

        $query = "UPDATE user SET inv_actv = inv_actv + 1 WHERE id = :user_id";
        db_query_execute($config, $query, convert_null_array(array(":user_id" => $invitation_record['user'])));
    }

    // Update accounts created stats counter
    // Also count this as a login

    update_account_created_stats($config, $login_region, $proxy_id, $user_group);
    update_login_stats($config, $login_region, $proxy_id, $user_group);

    if (strlen($_POST[create_account_email] ) && $require_email_verification)
    {
        // Send email verification email
        // Note: same transaction issues here as above.
        // TODO: allow user to force re-issue of verification email without changing
        // email address?  profile.php doesn't do this right now, so there's no way
        // to recover if the new account is created but the email doesn't get enqueued.
        // Update: way to recover is to set email to empty in profile, then
        // (re)set email to actual email address.

        enqueue_email_candidate_email(
            $config,
            $_POST[create_account_username],
            $proxy_id,
            $record_user[last_lang],
            $_POST[create_account_email],
            $email_candidate_token);

        // tell the user to check inbox for verification email

        $processing_results[] = ___("You have provided a new e-mail address which has not yet been verified.  Psiphon has sent a verification e-mail to this address.  To verify your address, click the link provided when you receive the e-mail.");
    }
    elseif ($_POST[create_account_email])
    {
        // Send a welcome email to the user's specified email address
        // Since email verification isn't required, this is sent immediately.
        // Note: same transaction issues here as above, both for enqueueing
        // the email and updating the welcome_email_sent flag
        // TODO: maybe a more robust way to ensure welcome emails are sent is
        // to implement a send_welcome_emails cron job which sends to all
        // user accounts with email not NULL?

        enqueue_welcome_email($config, $user_id);
    }

    // link for home page (https://proxy_ip/001/)
    $create_account_url = make_home_url($config, $proxy_id);

    break;
}
?>
<?/*
 *
 * Presentation Layer
 *
 */?>
<?=header("Content-type: text/html; charset=".charset());?>
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" "http://www.w3.org/TR/html4/loose.dtd">
<html>
    <head>
        <title><?=___("Welcome")?></title>
        <?=echo_css()?>
    </head>
    <body style="direction: <?=direction()?>">
        <table><tr>
            <td style="width:50%">
<?php
    include($_SERVER[DOCUMENT_ROOT]."/images/psiphonlogo.html");
?>
            </td>
            <td><font class="eighteenbluebold"><?=___("Welcome to Psiphon!")?></font></td>
        </tr></table>
        <?
        // When account created successfully, just output a short message and a link
        if ($create_account_url)
        {
            ?>
            <table><tr><td style="width:50em"><font class="fourteenblue">
            <?=sprintf(___("Welcome to your Psiphon account!<br>
<br>
Do not share this link or your Psiphon account login details on any public forum.  This will help protect Psiphon from being blocked in your region.<br>
<br>
You may encounter an \"invalid security certificate\" warning when you activate or access your Psiphon account.  Please select \"continue\" or \"add an exception\" to enable your browser to establish a secure connection to Psiphon.<br>
<br>
You must use the following link to login to Psiphon: <br>
<br>
%s<br>
<br>
The Psiphon Team<br>"),"<a href=\"{$create_account_url}\">{$create_account_url}</a>\n")?>
            </font></td></tr></table>
            </body></html>
            <?
            exit(0);
        }
        // Otherwise, output the following form
        ?>
        <form name="create_account" method="post" action="<?=getenv("REQUEST_URI")?>">
            <input name="csrf_token" value="<?=$record_user['csrf_token']?>" type="hidden">
            <input name="create_account_submit" value="1" type="hidden">
            <input name="create_account_lang" value="<?=$record_user["last_lang"]?>" type="hidden">
            <br>
            <table cellpadding="4" cellspacing="0">

                <tr><td><font class="sixteenblue"><div style="width:50em;"><?=___("Complete this form to create your Psiphon account.")?><br><br><?=$privacy_info?></div></font></td></tr>
            
                <tr><td><?=echo_processing_results_and_errors();?></td></tr>

                <table cellpadding="4" cellspacing="0">
                    <tr>
                        <td><font class="fourteenblue"><?=___("Username")?></font></td>
                        <td><input type="text" name="create_account_username" value="<?=escape_html($_POST[create_account_username])?>" size="35" maxlength="32"></td>
                    </tr>
                    <tr>
                        <td><font class="fourteenblue"><?=___("Password")?></font></td>
                        <td><input type="password" name="create_account_password" size="35"></td>
                    </tr>
                    <tr>
                        <td><font class="fourteenblue"><?=___("Verify password")?></font></td>
                        <td><input type="password" name="create_account_verify_password" size="35"></td>
                    </tr>
                    <tr>
                        <td><font class="fourteenblue"><?=___("E-mail address (optional)")?></font></td>
                        <td><input type="text" name="create_account_email" value="<?=escape_html($_POST[create_account_email] ? $_POST[create_account_email] : $default_email)?>" size="35" maxlength="64"></td>
                    </tr>
                    <?
                    // In invitation mode, ask for invite code; other modes check CAPCHA
                    if ($create_account_mode == create_account_mode_invitation) {?>
                        <tr>
                            <td><font class="fourteenblue"><?=___("Invitation code")?></font></td>
                            <td>
                                <? if ($invitation_code_readonly) { ?>
                                    <input name="create_account_invitation_code" value="<?=escape_html($_POST[create_account_invitation_code])?>" type="hidden">
                                    <?=escape_html($_POST[create_account_invitation_code])?>
                                <? } else { ?>
                                    <input type="text" name="create_account_invitation_code" value="<?=escape_html($_POST[create_account_invitation_code])?>" size="35" maxlength="64">
                                <? } ?>
                            </td>
                        </tr>
                    <?} else {?>
                        <tr>
                            <td valign="top"><font class="fourteenblue"><?=___("Security code")?></font></td>
                            <td><input type="text" name="create_account_security_code" value="" size="15" maxlength="4">
                            <br/>
                            <?/* image-code generates a CAPTCHA which is associated with a PHP session (e.g., _SESSION) */?>
                            <img src="image-code.php?<?=strtoupper(bin2hex(secure_rand(8)))?>" border="0" hspace="0" vspace="2" alt=""></td>
                        </tr>
                    <?}?>
                    <tr>
                        <td>&nbsp;</td>
                        <td><input type="submit" value="<?=___("Create account")?>">
                    </tr>
                </table>
            </table>
        </form>

        <br>
        <br>
        <?echo_language_buttons($config, $record_user, "create_account_lang");?>

        <script language="javascript" type="text/javascript">
            <!--
            document.<?=$curs?>.focus();
            //-->
        </script>
    </body>
</html>
