<?php
/*
    Psiphon Circumvention Platform
    Copyright (C) 2010 Psiphon Inc.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

function lookup_proxy($config, $proxy_id)
{
    $query = "SELECT * FROM proxy WHERE id = :proxy_id";
    $result = db_query_execute($config, $query, convert_null_array(array(":proxy_id" => $proxy_id)));

    $proxy_record = db_fetch_result($config, $result);

    return $proxy_record;
}

// NULL means infinite.
function current_invite_token_count_sql($config)
{
    // NOTE: The r-values here don't need to be parameterized because they aren't
    // user input (even second-order) -- they come from the $config array.

    $invite_token_sql = " (CASE user.grp ";

    foreach ($config['grp_invite_token_settings'] as $grp_id => $grp_invite_token_settings)
    {
        if (!user_class_has_permission($config, $grp_id, "invite"))
        {
            $invite_token_sql .= "WHEN {$grp_id} THEN 0 ";
            continue;
        }

        if ($grp_invite_token_settings['infinite_tokens'])
        {
            $invite_token_sql .= "WHEN {$grp_id} THEN NULL ";
            continue;
        }

        // Sanity check on config settings
        if (($grp_invite_token_settings['token_accrual_period_in_seconds'] < 1) ||
            ($grp_invite_token_settings['token_accrual_rate_per_period'] <= 0) ||
            ($grp_invite_token_settings['max_tokens'] <= 0) ||
            ($grp_invite_token_settings['delay_until_accrual_starts_in_seconds'] < 0)) // may be zero
        {
            include($_SERVER[DOCUMENT_ROOT]."/http-errors/500.php");
        }

        /*
         * Here's the math:
         * accrued_tokens := floor((CurrentTime() - last_token_accrual_time) / AccrualPeriod()) * AccrualRate();
         * current_tokens := min(MaxTokens(), last_token_count + accrued_tokens);
         *
         * last_token_count := current_tokens - spent_tokens;
         * last_token_accrual_time := last_token_accrual_time + floor((CurrentTime() - last_token_accrual_time) / AccrualPeriod()) * AccrualPeriod();
         */

        // Condition for this group ID
        $invite_token_sql .= "WHEN {$grp_id} THEN ";

        // If there's no accrual start time in the DB yet, we'll use user.created+delay.
        // But if that's after now, we'll just use now (to avoid negative numbers).
        $accrual_start_time = " (IFNULL(user.last_token_accrual_time, LEAST(UNIX_TIMESTAMP(), (user.created + {$grp_invite_token_settings['delay_until_accrual_starts_in_seconds']})))) ";

        // We want the number of whole accrual periods passed.
        $accrual_periods_passed = " (FLOOR((UNIX_TIMESTAMP() - {$accrual_start_time}) / {$grp_invite_token_settings['token_accrual_period_in_seconds']})) ";

        // Tokens accrued since last DB update.
        $accrued_tokens = " ({$grp_invite_token_settings['token_accrual_rate_per_period']} * {$accrual_periods_passed}) ";

        // The current tokens, limited by max.
        $current_tokens = " (LEAST({$grp_invite_token_settings['max_tokens']}, (user.last_token_count + {$accrued_tokens}))) ";

        $invite_token_sql .= $current_tokens;
    }

    $invite_token_sql .= "ELSE NULL END) ";

    return $invite_token_sql;
}

// The field name is invite_token_count. NULL means infinite.
function current_invite_token_count_select_sql($config)
{
    return current_invite_token_count_sql($config)." AS invite_token_count ";
}

function last_token_count_sql($config, $spent_tokens)
{
    // HACK: If the user has infinite tokens, then current_invite_token_count_sql will evaluate to NULL.
    //       But if we put NULL in the database, that will cause havoc in the future. So we'll just set
    //       user.last_token_count to zero -- it shouldn't matter if it's an infinite-tokens user.
    return " ( IFNULL((".current_invite_token_count_sql($config)." - ".$spent_tokens."), 0) ) ";
}

function last_token_accrual_time_sql($config, $user_record)
{
    $grp_invite_token_settings = $config['grp_invite_token_settings'][$user_record['grp']];

    // If last_token_accrual_time is null, then we need to calculate the start time.
    $accrual_start_time =
        " IFNULL(".
            "user.last_token_accrual_time, ".
            "LEAST(UNIX_TIMESTAMP(), (user.created + {$grp_invite_token_settings['delay_until_accrual_starts_in_seconds']}))) ";

    // We want the number of whole accrual periods passed.
    $accrual_periods_passed = " (FLOOR((UNIX_TIMESTAMP() - {$accrual_start_time}) / {$grp_invite_token_settings['token_accrual_period_in_seconds']})) ";

    // Calculate the time passed from the periods passed.
    $accrual_time_passed = " ({$accrual_periods_passed} * {$grp_invite_token_settings['token_accrual_period_in_seconds']}) ";

    $last_token_accrual_time = " ({$accrual_start_time} + {$accrual_time_passed}) ";

    return $last_token_accrual_time;
}

// Returns ID of proxy with IP of where the user is coming from. If not found, returns null.
function lookup_current_proxy($config)
{
    $query = "SELECT id FROM proxy WHERE hostname = :hostname";

    $result = db_query_execute($config, $query, convert_null_array(array(":hostname" => getenv("HTTP_HOST"))));

    if (!($proxy_id = db_fetch_result($config, $result)))
    {
        // not a known proxy
        return Null;
    }

    return $proxy_id['id'];
}

// $where_clause is SQL. $where_params is the array of params for $where_clause.
// Returns false if not found.
function lookup_user($config, $where_clause, $where_params)
{
    $query = "SELECT *, ".current_invite_token_count_select_sql($config)." FROM user WHERE (".$where_clause.")";

    $result = db_query_execute($config, $query, convert_null_array($where_params));
    
    $user_record = db_fetch_result($config, $result);

    if (!$user_record)
    {
        return false;
    }

    // Get the user's assigned proxies

    $query = "SELECT proxyid FROM userproxy WHERE userid = :userid";

    $result = db_query_execute($config, $query, convert_null_array(array(":userid" => $user_record['id'])));

    $user_record['assigned_proxies'] = array();

    while ($record = db_fetch_result($config, $result))
    {
        array_push($user_record['assigned_proxies'], $record['proxyid']);
    }

    return $user_record;
}

/**
 * Deletes users and all related data
 *
 * @param  $config
 * @param  $query_affected_user_ids_from_part           SQL conditions to select affected user ids with placeholders
 * @param  $param_array                                 placeholder values
 * @return bool                                         true on success, false on error
 */
function delete_user_and_related_data($config, $query_affected_user_ids_from_part, Array $param_array) {
    //   We don't know how big the user id list could be; So, we will use subselect here (Rod's recommendation)
    $query_affected_user_ids = 'SELECT user.id ' . $query_affected_user_ids_from_part;

    $query_delete_bookmark = "DELETE FROM bookmark WHERE user IN (" . $query_affected_user_ids . ")";
    $query_delete_cookie = "DELETE FROM cookie WHERE user IN (" . $query_affected_user_ids . ")";
    $query_delete_invitation = "DELETE FROM invitation WHERE user IN (" . $query_affected_user_ids . ")";
    $query_delete_ticket = "DELETE FROM ticket WHERE user IN (" . $query_affected_user_ids . ")";
    $query_delete_userproxy = "DELETE FROM userproxy WHERE userid IN (" . $query_affected_user_ids . ")";
    // Delete from user table only after deleting all other data
    $query_delete_user = "DELETE " . $query_affected_user_ids_from_part;

    if (
            !db_query_execute($config, $query_delete_bookmark, convert_null_array($param_array), true) ||
            !db_query_execute($config, $query_delete_cookie, convert_null_array($param_array), true) ||
            !db_query_execute($config, $query_delete_invitation, convert_null_array($param_array), true) ||
            !db_query_execute($config, $query_delete_ticket, convert_null_array($param_array), true) ||
            !db_query_execute($config, $query_delete_userproxy, convert_null_array($param_array), true) ||
            !db_query_execute($config, $query_delete_user, convert_null_array($param_array), true)
    ) {
        return false;
    }

    return true;
}

/* Turns a (comma-separated, usually) list of values into a values that can be used in a
 * parameterized SQL query.
 *
 * INPUT:
 *      $string_list: Looks like "x,y,z"
 *      $separator: "," or ", " or whatever
 *      $param_prefix: The prefix of the param names that will be used --
 *          doesn't matter what it is, as long as it's unique in the query.
 * OUTPUT:
 *      $param_string: Looks like "NULL,:param1,:param2,:param3" -- used in the query string (between parentheses).
 *          Note that the NULL is just so things work okay if there are no list values.
 *      $param_array: Looks like array(":param1" => "x", ""param2" => "y", etc.) -- passed as params to execute
 * RETURN:
 *      none
 */
function string_list_to_params($string_list, $separator, $param_prefix, &$param_string, &$param_array)
{
    $param_string = "NULL";
    $param_array = array();

    if (substr($param_prefix, 0, 1) != ":")
    {
        $param_prefix = ":".$param_prefix;
    }

    $list_array = explode($separator, $string_list);

    $count = 1;
    foreach ($list_array as $list_item)
    {
        $param_name = $param_prefix.$count++;

        $param_string .= ",".$param_name;
        $param_array[$param_name] = $list_item;
    }
}

// $changes_array should be an array with elements 'fieldname' => value (value can be Null)
// Input/output params are appended to.
function changes_array_to_sql_and_params_for_insert(
            $changes_array,
            &$fields_sql,
            &$values_sql,
            &$param_array)
{
    $fields = '';
    $values = '';
    $params = array();

    foreach ($changes_array as $field => $value)
    {
        if (strlen($fields))
        {
            $fields .= ',';
            $values .= ',';
        }

        // To ensure the parameter name is unique, we'll add a random number to the end.
        $fieldname = ':'.$field.rand(10, 99);

        $fields .= $field;

        if ($value === Null)
        {
            $values .= 'NULL';
        }
        else
        {
            $values .= $fieldname;
            $params[$fieldname] = $value;
        }
    }

    $fields_sql .= $fields;
    $values_sql .= $values;
    $param_array = array_merge($param_array, $params);
}

// $changes_array should be an array with elements 'fieldname' => value (value can be Null)
// Input/output params are appended to.
function changes_array_to_sql_and_params_for_update(
            $changes_array,
            &$update_sql,
            &$param_array)
{
    $sql = '';
    $params = array();

    foreach ($changes_array as $field => $value)
    {
        if (strlen($sql))
        {
            $sql .= ',';
        }

        // To ensure the parameter name is unique, we'll add a random number to the end.
        $fieldname = ':'.$field.rand(10, 99);
        // The field might be table.column, so remove any periods from the fieldname
        $fieldname = str_replace('.', '', $fieldname);

        if ($value === Null)
        {
            $sql .= $field.'=NULL';
        }
        else
        {
            $sql .= $field.'='.$fieldname;
            $params[$fieldname] = $value;
        }
    }

    $update_sql .= $sql;
    $param_array = array_merge($param_array, $params);
}

// Returns true if a record exists in $table where $column is equal to $value
// in the space restricted by $where_clause.
function db_check_if_value_exists(
            $config,
            $table,
            $column,
            $value,
            $where_clause = "",
            $where_clause_params = array())
{
    $query = "SELECT COUNT(*) AS count FROM $table WHERE $column = :db_check_if_value_exists_value";

    if (strlen($where_clause) > 0)
    {
        $query .= " AND ($where_clause)";
    }

    $result = db_query_execute(
                $config,
                $query,
                convert_null_array(
                    array(":db_check_if_value_exists_value" => $value)
                    + $where_clause_params));
            
    $record = db_fetch_result($config, $result);

    return ($record['count'] > 0);
}

?>
