<?
/*
    Psiphon Circumvention Platform
    Copyright (C) 2009 Psiphon Inc.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


include($_SERVER[DOCUMENT_ROOT]."/includes/sql_error.php");

function db_connect($config)
{
    try
    {
        $pdo_connection = new PDO('mysql:host='.$config['db_host'].';dbname='.$config['db_name'], $config['db_user'], $config['db_pass']);
    }
    catch (PDOException $e)
    {
        sql_error($config, "Connect failed", $e->getMessage(), __FUNCTION__, __LINE__);
    }

    return($pdo_connection);
}

/*
 * $params must look like array(":var1"=>$var1, ":var2"=>$var2, ...).
 * $return_on_error should be false for the web-php usages, but true for the CLI-php (cron jobs)
 *
 * RETURN: - prepared and executed statement on success (which is ready for result binding and fetching or whatever)
 *         - if $return_on_error is true, returns false on error; otherwise 503 and exit(0) will result 
 */
function db_query_execute($config, $sql, $params, $return_on_error = false)
{
    /*
     * Initially there was an option to specify "driver options' for prepare(),
     * but I found that scrollable cursors didn't work as I expected, so I removed
     * the option (i.e., once the cursor reached the last result, you can't scroll anymore).
     */

    if (false) // DEBUG: Used for doing DB call coverage tests
    {
        $trace = debug_backtrace();
        $location = substr($trace[0]['file'], strrpos($trace[0]['file'], "/")+1).":".$trace[0]['line'];

        $current_file = substr(__FILE__, strrpos(__FILE__, "/")+1).":";
        if (strncmp($location, $current_file, strlen($current_file)) != 0)
        {
            db_query_execute(
                $config,
                "INSERT IGNORE INTO db_calls (location, count) VALUES (:location, 1) ".
                "ON DUPLICATE KEY UPDATE count = count+1",
                convert_null_array(array(":location" => $location)));
        }
    }

    try
    {
        $stmt = $config['sql']->prepare($sql);
    }
    catch (PDOException $e)
    {
        if (!$return_on_error)
        {
            // sql_error includes 503.php, which exits
            sql_error($config, $sql, $e->getMessage(), __FUNCTION__, __LINE__);
        }
        return false;
    }

    if (!$stmt->execute($params))
    {
        if (!$return_on_error)
        {
            // sql_error includes 503.php, which exits
            sql_error($config, $sql, implode(' // ', $stmt->errorInfo()), __FUNCTION__, __LINE__);
        }
        return false;
    }

    return $stmt;
}

/*
 * $stmt must be the result of a successful db_query_execute() call.
 *
 * RETURN: - current result record array on success
 *         - returns false if no more results
 */
function db_fetch_result($config, $stmt, $return_on_error = false)
{
    if (!$stmt)
    {
        if (!$return_on_error)
        {
            // sql_error includes 503.php, which exits
            sql_error($config, "Null stmt", "", __FUNCTION__, __LINE__);
        }
        
        return false;
    }

    $result = $stmt->fetch(PDO::FETCH_ASSOC);

    return $result;
}

/*
 *  For INSERT, UPDATE, DELETE queries.
 *
 * $stmt must be the result of a successful db_query_execute() call.
 */
function db_num_rows_affected($stmt)
{
    if (!$stmt)
    {
        return $stmt;
    }

    return $stmt->rowCount();
}

/*
 * Get the auto-increment ID value of the last row inserted.
 */
function db_get_last_insert_id($config)
{
    return $config['sql']->lastInsertId();
}

// Returns true on success or false on failure.
function db_free_result($stmt)
{
    if (!$stmt)
    {
        return false;
    }

    return $stmt->closeCursor();
}

/*
 * This function is used to ease the transition between the old non-parameterized
 * SQL code and the new parameterized SQL code.
 * In the old code, there was a lot of stuff that looked like:
 *   $query = "UPDATE user SET region = '{$region}'";
 * In the new code that becomes:
 *   $query = "UPDATE user SET region = :region"; ...array(":region" => $region)...
 * But those aren't equivalent. If $region is Null, in the old code user.region would
 * get set to an empty string. In the new code, it will get set to NULL.
 *
 * convert_null_array changes any Null values in the $array_with_nulls to
 * empty strings.
 */
function convert_null_array($array_with_nulls)
{
    if (!is_array($array_with_nulls))
    {
        return $array_with_nulls;
    }

    foreach ($array_with_nulls as $key => $val)
    {
        if (is_null($val))
        {
            $array_with_nulls[$key] = "";
        }
    }

    return $array_with_nulls;
}
