<?php
/*
    Psiphon Circumvention Platform
    Copyright (C) 2009 Psiphon Inc.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 * This file contains functions (both php and javascript) that help the HTML
 * presentation code get generated and operate.
 */

// Output style sheet
function echo_css()
{
    global $record_user;
    echo "<style>\n";
    include($_SERVER[DOCUMENT_ROOT]."/css/psistyle.css");
    echo "</style>\n";
}

// Outputs a combo box control, built from an array of values and labels.
function echo_combo_from_array(
            $field_name,      // the name of the control
            $options,         // array of values and labels to put into combo
            $transform_func,  // Optional. The function that will be used to modify the label before it is shown.
            $selected_value,  // The value that should be selected in the combo by default.
            $add_empty_option = 1,  // If 1, a blank option will be added to the combo.
            $empty_option_text = "")// The text to show for the empty option
{
    // populate combo box with key/label results from $options

    echo "<select name=\"${field_name}\">\n";

    if ($add_empty_option == 1)
    {
        echo "<option value=\"\">{$empty_option_text}</option>\n";
    }

    foreach ($options as $value => $label)
    {
        echo "<option value=\"{$value}\"";

        // restore last proxy selection

        if (($selected_value !== Null) && ($selected_value == $value))
        {
            echo "selected=1";
        }

        if ($transform_func)
        {
            $label = $transform_func($label);
        }
        echo ">{$label}</option>\n";
    }

    echo "</select>\n";
}

// Outputs a combo box control, built from a SQL query.
function echo_combo_from_sql(
            $config,          // system config array
            $field_name,      // the name of the control
            $sql_query,       // the sql for the DB query that will provide the options
            $sql_query_params,// the params used in $sql_query
            $key_name,        // the name of the column returned from the DB query that will be the value for each option
            $label_name,      // the name of the column returned from the DB query that will be the label for each option
            $transform_func,  // Optional. The function that will be used to modify the label before it is shown.
            $selected_value,  // The value that should be selected in the combo by default.
            $add_empty_option = 1,  // If 1, a blank option will be added to the combo.
            $empty_option_text = "")// The text to show for the empty option
{
    // populate combo box with key/label results from $sql_query

    $result = db_query_execute($config, $sql_query, convert_null_array($sql_query_params));

    $options = array();

    while ($record = escape_html(db_fetch_result($config, $result)))
    {
        $options[$record[$key_name]] = $record[$label_name];
    }

    echo_combo_from_array(
            $field_name,
            $options,
            $transform_func,
            $selected_value,
            $add_empty_option,
            $empty_option_text);
}

function proxy_label_sql($proxy_table_name)
{
    return "CONCAT({$proxy_table_name}.name, ' - ', {$proxy_table_name}.hostname )";
}

// Outputs a combo box with available proxies.
function echo_proxy_combo(
            $config,                // system config array
            $field_name,            // the name of the control
            $selected_id,           // The value that should be selected in the combo by default.
            $add_empty_option,      // If 1, a blank option will be added to the combo.
            $empty_option_text = "")// The text to show for the empty option
{
    global $record_user;

    $query_proxy = "SELECT id, ".proxy_label_sql("proxy")." AS label FROM proxy";

    $query_proxy_params = array();
    
    if (!user_has_permission($config, $record_user, "other_nodes"))
    {
        $query_proxy .= " WHERE id IN (SELECT proxyid FROM userproxy WHERE userid = :userid) ";
        $query_proxy_params[":userid"] = $record_user['id'];
    }

    $query_proxy .= " ORDER BY 2";
    
    echo_combo_from_sql(
              $config,
              $field_name,
              $query_proxy,
              $query_proxy_params,
              "id",
              "label",
              NULL,
              $selected_id,
              $add_empty_option,
              $empty_option_text);
}

// Outputs a combo box with available languages.
function echo_lang_combo(
            $config,
            $field_name,            // the name of the control
            $selected_id,           // The value that should be selected in the combo by default.
            $add_empty_option = 1,  // If 1, a blank option will be added to the combo.
            $empty_option_text = "")// The text to show for the empty option
{
    $lang_combo = array();
    foreach ($config[lang] as $key => $val)
    {
        $lang_combo[$key] = ___($val["title"]);
    }

    echo_combo_from_array(
              $field_name,
              $lang_combo,
              NULL,
              $selected_id,
              $add_empty_option,
              $empty_option_text);
}

// Outputs a combo box with available countries (as derived from user.last_reg).
function echo_country_combo(
            $config,                // system config array
            $field_name,            // the name of the control
            $selected_id,           // The value that should be selected in the combo by default.
            $add_empty_option = 1,  // If 1, a blank option will be added to the combo.
            $empty_option_text = "")// The text to show for the empty option
{
    echo_combo_from_sql(
        $config,
        $field_name,
        "SELECT DISTINCT last_reg AS country FROM user WHERE LENGTH(last_reg) > 0 ORDER BY country",
        array(),
        "country",
        "country",
        Null,
        $selected_id,
        $add_empty_option,
        $empty_option_text);
}

// Outputs a combo box with user classes.
// Can be restricted to show only equal or lesser classes (than current user).
function echo_user_class_combo(
            $config,                     // system config array
            $field_name,                 // the name of the control
            $selected_class,             // The value that should be selected in the combo by default.
            $add_empty_option = 1,       // If 1, a blank option will be added to the combo.
            $only_managed_classes = 1)   // If 1, classes that the current user can't manage won't be shown.
{
    global $record_user;

    $user_classes = get_user_class_names($config, $record_user, $only_managed_classes, 1);

    echo_combo_from_array(
            $field_name,
            $user_classes,
            Null,   // no transform function
            $selected_class,
            $add_empty_option);
}

function echo_check_all_handler()
{
?>
    <script type="text/javascript">
    // Flip all checkboxes in specified form; also toggles button text and action.
    function checkAll(
                obj,                    // Object containing the checkboxes to toggle.
                checkbox_name_prefix,   // Prefix of the name of the checkboxes to toggle (all others will be skipped).
                button_id)              // ID of the button to change the text of.
    {
        var button = document.getElementById(button_id);
        var check_all_text = "<?=___("Check all")?>";
        var uncheck_all_text = "<?=___("Uncheck all")?>";

        for (i = 0; i < obj.elements.length; i++)
        {
            if (obj.elements[i].type == "checkbox" && obj.elements[i].name.substr(0, checkbox_name_prefix.length) == checkbox_name_prefix)
            {
                obj.elements[i].checked = (button.value == check_all_text ? true : false);
            }
        }
        button.value = (button.value == check_all_text ? uncheck_all_text : check_all_text);
    }
    </script>
<?
}


// Results array. Every element in this array is displayed at the top of the page.
// Pages should put processing result strings into this array.
$processing_results = array();

// Error array. Every element in this array is displayed at the top of the page in red.
// Pages should put processing error strings into this array.
$processing_errors = array();

// Should probably only be called within header.php
function echo_processing_results_and_errors()
{
    global $processing_results;
    global $processing_errors;

    if (count($processing_results) > 0 || count($processing_errors) > 0)
    {
        echo "<table cellpadding=\"2\" cellspacing=\"0\" >\n";

        if (count($processing_results) > 0)
        {
            echo "<tr><td>\n";
            echo "<table>\n";

            $counter = 0;
            foreach ($processing_results as $result)
            {
                echo "<tr><td id=\"processing_result[$counter]\" class=\"processing_result\">".$result."</td></tr>\n";
                $counter++;
            }

            echo "</table>\n";

            echo "</td></tr>\n";
        }

        if (count($processing_errors) > 0)
        {
            echo "<tr><td>\n";
            echo "<table>\n";

            $counter = 0;
            foreach ($processing_errors as $error)
            {
                echo "<tr><td id=\"processing_error[$counter]\" class=\"processing_error\">".$error."</td></tr>\n";
                $counter++;
            }

            echo "</table>\n";

            echo "</td></tr>\n";
        }

        echo "</table>\n";
        echo "<br/>\n";
    }
}

function echo_language_buttons(
            $config,
            $record_user,
            $language_variable_name,
            $row_width = 0
)
{
            echo "<table border=\"0\" cellpadding=\"2\" cellspacing=\"0\"><tr>";
            foreach ($config[lang] as $l_key => $l_val)
            {
                $action = escape_html(getenv("REQUEST_URI"));
                echo "<td><form name=\"{$l_key}_selection\" method=\"post\" action=\"{$action}\">";
                echo "<input type=\"hidden\" name=\"{$language_variable_name}\" value=\"{$l_key}\">";
                echo "<input name=\"csrf_token\" value=\"{$record_user['csrf_token']}\" type=\"hidden\">";
                // The button
                echo "<input type=\"submit\" value=\"{$l_val["translated_title"]}\" style=\"border-style:none;cursor:pointer;";
                echo ($record_user["last_lang"] == $l_key) ? "background-color:#BBBBBB;" : "background-color:#EEEEEE;";
                echo "font-family:Arial,Helvetica,sans-serif";
                echo "\">";

                echo "</form></td>";
                if ($row_width && !(++$count % $row_width))
                {
                    echo "</tr><tr>";
                }
            }
            echo "</tr></table>";
}

?>
