<?php
/*
    Psiphon Circumvention Platform
    Copyright (C) 2009 Psiphon Inc.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Returns the ID of the anonymous user class.
function get_anonymous_user_class_id()
{
    return "4";
}

// Returns the ID of the normal User user class.
function get_normal_user_class_id()
{
    return "3";
}

// Returns the ID of the Power User user class.
function get_power_user_class_id()
{
    return "2";
}

// Returns the ID of the Admin user class.
function get_admin_user_class_id()
{
    return "0";
}

// Returns the ID of the Propagator user class.
function get_propagator_user_class_id()
{
    return "6";
}

// Returns the ID of the Reporter user class.
function get_reporter_user_class_id()
{
    return "7";
}

// Returns 1 if user is of anonymous type, 0 otherwise.
function is_anonymous_user($record_user)
{
    if ($record_user["grp"] == get_anonymous_user_class_id())
    {
        return 1;
    }

    return 0;
}

// Returns the translated name of the user class.
function get_user_class_name($class_id)
{
    switch($class_id)
    {
        case 0:
            return ___("Admin");
        case 2:
            return ___("Power User");
        case 3:
            return ___("User");
        case 4:
            return ___("Guest User");
        case 6:
            return ___("Propagator");
        case 7:
            return ___("Reporter");
        default:
            return NULL;
    }
}

// Returns an array that maps user class IDs to the translated string representation of them.
// NOTE: Anonymous users should almost always be excluded.
// NOTE: If $only_managed_classes is 1 and $exclude_anonymous is 0, Anonymous Users will still be excluded if they're not managed.
// Results will be ordered from least privileged to most (using $config["grp_order"]).
function get_user_class_names($config, $record_user, $only_managed_classes, $exclude_anonymous = 1)
{
    /* The sorting explained:
     *  - First create an array of this form: (class_rank1 => class_id1, ...)
     *  - Then sort the array by key (class_rank).
     *  - Then extract from that the final array to return of the form: (class_id1 => class_string1, ...)
     */

    $user_classes_by_rank = array();

    foreach ($config["grp"] as $class_id => $class_permissions)
    {
        // If desired by caller, only show manageable classes.
        if ($only_managed_classes && !can_user_manage_class($config, $record_user, $class_id))
        {
            continue;
        }

        // if desired by caller, exclude the anonymous user class
        if ($exclude_anonymous && ($class_id == get_anonymous_user_class_id()))
        {
            continue;
        }

        // The sorting code uses $config["grp_order"] to determine where the current
        // class should end up in the output array.

        $current_class_rank = array_search($class_id, $config["grp_order"]);
        if (!$current_class_rank)
        {
            // Error! Need to update $config["grp_order"]?
            return array();
        }

        $user_classes_by_rank[$current_class_rank] = $class_id;
    }

    // Sort the array by rank.
    ksort($user_classes_by_rank);

    $user_class_names = array();

    foreach ($user_classes_by_rank as $class_rank => $class_id)
    {
        $user_class_names[$class_id] = get_user_class_name($class_id);
    }

    return $user_class_names;
}

function user_class_has_permission($config, $user_class_id, $permission_name)
{
    if ($config["grp"][$user_class_id] & $config["perm"][$permission_name])
    {
        return 1;
    }

    return 0;
}

// Returns 1 if user has the specified permission, 0 otherwise.
// $record_user must be an array with a "grp" element that is a class ID. For example, the global $record_user variable.
function user_has_permission($config, $record_user, $permission_name)
{
    if (user_class_has_permission($config, $record_user["grp"], $permission_name))
    {
        return 1;
    }

    return 0;
}

// Returns array of classes that have the given permission.
function get_user_classes_with_permission($config, $permission_name)
{
    $matching_classes = array();
    foreach ($config['grp'] as $user_class => $permissions)
    {
        if (user_class_has_permission($config, $user_class, $permission_name))
        {
            $matching_classes[] = $user_class;
        }
    }

    return $matching_classes;
}

// Returns 1 if user class $manager_class_id can manage user class $managed_class_id. Returns 0 otherwise.
function can_class_manage_class($config, $manager_class_id, $managed_class_id)
{
    $need_perm = "can_grant_grp_".$managed_class_id;
    
    return user_class_has_permission($config, $manager_class_id, $need_perm);
}

// Returns 1 if user can manage class.
function can_user_manage_class($config, $record_user, $class_id)
{
    return can_class_manage_class($config, $record_user["grp"], $class_id);
}

// Returns a SQL where clause that can be used to restrict a query to just managed users.
// If no users can be managed, " (0) " is returned.
function get_user_managed_classes_sql_condition($config, $record_user)
{
    $managed_classes = "";

    foreach ($config["grp"] as $class_id => $class_permissions)
    {
        if (!can_user_manage_class($config, $record_user, $class_id))
        {
            continue;
        }
        
        if (strlen($managed_classes) > 0)
        {
            $managed_classes .= ",";
        }

        $managed_classes .= $class_id;
    }

    if (strlen($managed_classes) > 0)
    {
        $sql_condition = " (user.grp in (".$managed_classes.")) ";
    }
    else
    {
        $sql_condition = " (0) ";
    }

    return $sql_condition;
}

// Returns TRUE if user has permission to access indicated proxy. False otherwise.
function can_current_user_access_proxy($config, $record_user, $proxyid)
{
    if (user_has_permission($config, $record_user, "other_nodes"))
    {
        return true;
    }

    $query = "SELECT COUNT(*) AS found FROM userproxy WHERE userid = :userid AND proxyid = :proxyid";


    $result = db_query_execute(
                $config,
                $query,
                convert_null_array(array(
                    ':userid' => $record_user['id'],
                    ':proxyid' => $proxyid)));

    $record = db_fetch_result($config, $result);

    return (int)$record['found'] > 0;
}

?>
